#include "global.h"

/******************************************************************************

	Welcome to the dcl module.  A parser of C declarations.

	This is what's called `a kludge'.
	However, based on the good foundaries of the other modules
	we can advance adding extensions here, making the code even
	more hairy.

	Eventually this becomes a parser for LWC declarations.

******************************************************************************/

recID current_scope [16];
int top_scope;

#define BASETYPE_ABSTRACT -2
#define BASETYPE_SKIP -3

typedef struct {
	int dclqual_i;
	Token dclqual [8];
	Token shortname [80];
	recID basetype;
	typeID basetype2;
	bool istypedef, isvirtual, isextern, isstatic, isauto,
	     isinline, isconst, islinkonce, ismodular;
	Token class_name;
	bool anon_union, maybe_ctor;
} Dclbase;

static Dclbase *outer_dclqual;

Token bt_macro;
Token bt_replace;

static NormPtr collect_dclqual (Dclbase *b, NormPtr p)
{
	while (ISDCLFLAG (CODE [p]))
		b->dclqual_i |= 1 << (CODE [p++] - RESERVED_auto);
	return p;
}

/* Take the declaration qualifiers from a decl and
 * enable them for another decl */
static void append_dclqual (Dclbase *b, Dclbase *m)
{
	if (b->isstatic) {
		if (!intchr (m->dclqual, RESERVED_extern)
		&& !intchr (m->dclqual, RESERVED_static))
			intcatc (m->dclqual, RESERVED_static);
		m->isstatic = 1;
	}
	if (b->isinline) {
		if (!intchr (m->dclqual, RESERVED_inline))
			intcatc (m->dclqual, RESERVED_inline);
		m->isinline = 1;
	}
	if (b->isvirtual) m->isvirtual = 1;
	if (b->isauto) m->isauto = 1;
	if (b->ismodular) m->ismodular = 1;
}

static void sumup_dclqual (Dclbase *b)
{
#define ADDFLAG(x) b->is ## x = b->dclqual_i & (1 << (RESERVED_ ## x - RESERVED_auto))
	int i, j = 0;
	for (i = RESERVED_auto; ISDCLFLAG (i); i++)
		if ((b->dclqual_i & (1 << (i - RESERVED_auto))))
			if (i != RESERVED_virtual && i != RESERVED_auto
			&& i != RESERVED_linkonce && i != RESERVED_modular)
				b->dclqual [j++] = i;
	b->dclqual [j] = -1;
	ADDFLAG(typedef);	ADDFLAG(virtual);
	ADDFLAG(extern);	ADDFLAG(static);
	ADDFLAG(auto);		ADDFLAG(inline);
	ADDFLAG(const);		ADDFLAG(linkonce);
	ADDFLAG(modular);
}

static NormPtr bt_builtin (Dclbase *b, NormPtr p)
{
	Token bt = RESERVED_int;
	int sh, lo, si, us, i = 0;

	for (sh = lo = si = us = 0; ISTBASETYPE (CODE [p]); p++)
		switch (CODE [p]) {
		 case RESERVED_short: sh++;
		ncase RESERVED_long: lo++;
		ncase RESERVED_signed: si++;
		ncase RESERVED_unsigned: us++;
		ndefault: bt = CODE [p];
		}

	switch (bt) {
	 case RESERVED_float:
		 b->basetype = B_FLOAT;
		 b->shortname [i++] = RESERVED_float;
	ncase RESERVED_void:
		 b->basetype = B_VOID;
		 b->shortname [i++] = RESERVED_void;
	ncase RESERVED_double:
		if (lo) {
		 b->basetype = B_LDOUBLE;
		 b->shortname [i++] = RESERVED_long;
		 b->shortname [i++] = RESERVED_double;
		} else {
		 b->basetype = B_DOUBLE;
		 b->shortname [i++] = RESERVED_double;
		}
	ncase RESERVED_char:
		if (us) {
		 b->basetype = B_UCHAR;
		 b->shortname [i++] = RESERVED_unsigned;
		} else
		 b->basetype = B_SCHAR;
		b->shortname [i++] = RESERVED_char;
	ncase RESERVED_int:
		if (lo == 2) {
		 if (us) {
		  b->basetype = B_ULONG;
		  b->shortname [i++] = RESERVED_unsigned;
		 } else
		  b->basetype = B_SLONG;
		 b->shortname [i++] = RESERVED_long;
		} else if (sh) {
		 if (us) {
		  b->basetype = B_USINT;
		  b->shortname [i++] = RESERVED_unsigned;
		 } else
		  b->basetype = B_SSINT;
		 b->shortname [i++] = RESERVED_short;
		 b->shortname [i++] = RESERVED_int;
		} else {
		 if (us) {
		  b->basetype = B_UINT;
		  b->shortname [i++] = RESERVED_unsigned;
		 } else
		  b->basetype = B_SINT;
		 b->shortname [i++] = RESERVED_int;
		}
		if (lo) b->shortname [i++] = RESERVED_long;
	ndefault:
		parse_error (p, "BUG");
	}
	b->shortname [i] = -1;
	return p;
}

static recID is_object_in_global (NormPtr p)
{
	while (CODE [p] != '.' && CODE [p] != '(' && CODE [p] != ';' && CODE [p] != ',')
		p++;
	if (CODE [p] != '.') return 0;
	return lookup_struct (CODE [p - 1]);
}

static NormPtr bt_name (Dclbase *b, NormPtr p)
{
	recID r;
	Token x, y, t;

	if ((x = CODE [p]) == RESERVED__CLASS_)
		x = name_of_struct (current_scope [top_scope]);

	redo: switch (y = is_typename (x)) {
	case 0:
		/* Feature: template abstract classes */
		if (bt_macro && x == bt_macro) {
			x = bt_replace;
			goto renamed;
		}
		/* Feature: local typedefs in global scope */
		if (top_scope) {
			int i;
			for (i = top_scope; i; --i)
				if (t = lookup_local_typedef (current_scope [i], x)) {
					x = t;
					goto redo;
				}
		}
		if ((r = is_object_in_global (p)))
			if (x = lookup_local_typedef (r, x))
				goto redo;
		/* - - - - - - - - - - - - - - - - - - - - */
		if (is_extern_templ_func (&p)) {
			b->basetype = BASETYPE_SKIP;
			return p;
		}
		/* ////////////////////////////////// */
		parse_error_pt (p, x, "Not good for a base type");
	ncase 1:
		b->basetype = -1;
		b->basetype2 = lookup_typedef (x);
		b->shortname [0] = x;
		b->shortname [1] = -1;
	ncase 2:
	renamed:
		/* Feature: structure.constructor not a base type */
		if (CODE [p + 1] == '.') {
			Token t = CODE [p + 2];
			if (t == RESERVED_ctor
			|| t == RESERVED_dtor || t == CODE [p]
			|| (t == '~' && CODE [p + 3] == CODE [p])) {
				b->basetype = B_VOID;
				b->shortname [0] = -1;
				return p;
			}
		}
		/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
		b->basetype = lookup_struct (x) ?: x;
		if (lookup_struct (x) && is_aliasclass (b->basetype))
			sintprintf (b->shortname, x, -1);
		else	sintprintf (b->shortname, RESERVED_struct, x, -1);
		b->maybe_ctor = true;
	ncase 3:
		b->basetype = B_SINT;
		b->shortname [0] = RESERVED_enum;
		b->shortname [1] = x;
		b->shortname [2] = -1;
	ndefault:
		/* local typedef redirection */
		/* XXXX: could we loop ??? */
		x = y;
		goto redo;
	}
	return p + 1;
}

/* Feature: do typeof */
static NormPtr bt_typeof (Dclbase *b, NormPtr p)
{
	exprret E;

	if (CODE [p++] != '(')
		parse_error (p, "typeof '('");
	p = parse_expression (p, &E, NORMAL_EXPR);
	if (!E.ok) parse_error (p, "compilation halted due to expression error");
	if (CODE [p++] != ')')
		parse_error (p, "typeof (...')'");

	sintprintf (b->shortname, RESERVED___typeof__, '(', ISTR (E.newExpr), ')', -1);
	b->basetype = -1;
	b->basetype2 = E.type;

	return p;
}
/* ~~~~~~~~~~~~~~~~~ */

static NormPtr enum_dcl (Dclbase *b, NormPtr p, bool benum)
{
	int i = 0;

	enter_enum (b->shortname [1]);
	outprintf (GLOBAL, ISTR (b->shortname), '{', -1);
	for (;; p++) {
		if (CODE [p] == '}') break;
		if (!ISSYMBOL (CODE [p]))
			parse_error (p, "enum noise internal");
		enter_enumconst (CODE [p]);
		output_itoken (GLOBAL, CODE [p++]);
		if (benum) {
			outprintf (GLOBAL, '=', binshift [31 & i++], -1);
		} else {
			if (CODE [p] == '=') {
				exprret E;
				p = parse_const_expression (1 + p, &E);
				if (E.ok)
				outprintf (GLOBAL, '=', ISTR (E.newExpr), - 1);
			}
		}

		if (CODE [p] == ',') {
			output_itoken (GLOBAL, ',');
			continue;
		}

		if (CODE [p] == '}')
			break;
		parse_error_ll ("error in enum");
	}
	outprintf (GLOBAL, '}', ';', -1);
	return p + 1;
}

static NormPtr bt_enum (Dclbase *b, NormPtr p, bool benum)
{
	b->basetype = B_SINT;
	b->shortname [0] = RESERVED_enum;
	b->shortname [2] = -1;
	if (CODE [p] == '{') {
		b->shortname [1] = name_anonymous_enum ();
		return enum_dcl (b, p + 1, benum);
	}
	if (!ISSYMBOL (CODE [p]))
		parse_error_ll ("erroneous enum declaration");
	b->shortname [1] = CODE [p++];
	if (CODE [p] == '{')
		return enum_dcl (b, p + 1, benum);
	return p;
}

static NormPtr parse_structure (Token, Dclbase*, NormPtr, Token, Token*);
static NormPtr parse_inheritance (NormPtr, recID, Token[]);

static NormPtr bt_struct (Dclbase *b, NormPtr p, Token struct_or_union)
{
	Token abstract_parents [64] = { -1, };

	bool anon = CODE [p] == '{' || CODE [p] == ':';

	if (!anon && !ISSYMBOL (CODE [p]))
		parse_error_ll ("Das is verbotten");
	Token name = anon ? name_anonymous_struct () : CODE [p++];

	if (anon && struct_or_union == RESERVED_union)
		b->anon_union = true;

	bool need_fwd = !lookup_struct (name);

	if (have_abstract (name))
		parse_error_tok (name, "already have abstract class named that");

	b->basetype = enter_struct (name, struct_or_union == RESERVED_struct);

	/* Feature: Inheritance */
	if (CODE [p] == ':')
		p = parse_inheritance (p + 1, b->basetype, abstract_parents);
	/* # # # # # # # # # # #*/

	if (CODE [p] == '{')
		p = parse_structure (name, b, p + 1, struct_or_union, abstract_parents);

	sintprintf (b->shortname, is_aliasclass (b->basetype) ? BLANKT : struct_or_union,
		    name, -1);

	if (need_fwd && !is_aliasclass (b->basetype))
		outprintf (GLOBAL, ISTR (b->shortname), ';', -1);

	return p;
}

static Token specialized (Token, NormPtr);

/* Feature: Unique Specialization for template class */
static NormPtr bt_specialize (Dclbase *b, NormPtr p)
{
	Token c = CODE [p++], sp;

	if (!ISSYMBOL (c) || !have_abstract (c) || CODE [p] != '{')
		parse_error (p, "No such abstract class");

	/* * * * * * * * * * * */
	sp = specialized (c, ++p);
	/* * * * * * * * * * * */

	b->basetype = enter_struct (sp, true);

	if (abstract_has_special (c, b->basetype))
		p = skip_braces (p);
	else {
		Token abp [] = { c, -1 };
		outprintf (GLOBAL, iRESERVED_struct (b->basetype), sp, ';', -1);
		SAVE_VARC (b->isstatic, 0, s);
		p = parse_structure (sp, b, p, RESERVED_struct, abp);
		RESTOR_VARC (b->isstatic, s);
	}
	sintprintf (b->shortname, iRESERVED_struct (b->basetype), sp, -1);

	return p;
}

/* Feature: Base type abstract class */
static NormPtr bt_abstract (Dclbase *b, NormPtr p)
{
	Token an, par [32];
	int pc = 0;
	NormPtr p1;

	if (!ISSYMBOL (CODE [p]))
		parse_error (p, "anonymous abstract not possible");

	an = CODE [p++];
	if (CODE [p] == ':') {
		++p;
		do if (!issymbol (par [pc++] = CODE [p++]))
			parse_error (p, "Symbol expected");
		while (CODE [p++] == ',');
		--p;
	}
	par [pc] = -1;
	if (CODE [p++] != '{')
		parse_error (p, "abstract '{'");
	p = skip_braces (p1 = p);
	enter_abstract (an, par, p1);
	b->basetype = BASETYPE_ABSTRACT;
	return p;
}
/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

/* Feature: RegExp as a basetype. send to regexp.c */
static NormPtr bt_regexp (Dclbase *b, NormPtr p)
{
	b->basetype = BASETYPE_SKIP;
	return parse_RegExp (p, (b->isstatic!=0) + 2 * (b->isextern!=0) + 4 * (b->isinline!=0));
}

static NormPtr parse_basetype (Dclbase *b, NormPtr p)
{
	b->dclqual_i = 0;
	b->anon_union = false;
	b->maybe_ctor = false;
	p = collect_dclqual (b, p);
	sumup_dclqual (b);

	/* Feature : classes */
	if (CODE [p] == RESERVED_class)
		CODE [p] = RESERVED_struct;

	if (CODE [p] == RESERVED_ctor)
		b->basetype = B_VOID, b->shortname [0] = -1;
	else if (CODE [p] == RESERVED_dtor)
		b->basetype = typeID_voidP, b->shortname [0] = -1;
	else if (CODE [p] == RESERVED_enum)
		p = bt_enum (b, p + 1, false);
	else if (CODE [p] == RESERVED_benum)
		p = bt_enum (b, p + 1, true);
	else if (CODE [p] == RESERVED_struct || CODE [p] == RESERVED_union)
		p = bt_struct (b, p + 1, CODE [p]);
	else if (CODE [p] == RESERVED_template)	/* XXX next word can be only class */
		p = bt_abstract (b, p + 2);
	else if (CODE [p] == RESERVED_specialize)
		p = bt_specialize (b, p + 1);
	else if (CODE [p] == RESERVED_typeof)
		p = bt_typeof (b, p + 1);
	else if (CODE [p] == RESERVED_RegExp)
		p = bt_regexp (b, p + 1);
	else if (ISTBASETYPE (CODE [p]))
		p = bt_builtin (b, p);
	else if (ISSYMBOL (CODE [p]))
		/* Feature: constructor with class name */
		if ((CODE [p] == b->class_name || CODE [p] == bt_macro)
		&& CODE [p + 1] == '(') {
			b->basetype = B_VOID;
			b->shortname [0] = -1;
			CODE [p] = RESERVED_ctor;
		/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
		} else p = bt_name (b, p);
	else if (CODE [p] == '~' && CODE [p + 1] == b->class_name) {
	     /* Feature: ~obj  is a dtor */
		b->basetype = typeID_voidP;
		b->shortname [0] = -1;
		CODE [p + 1] = RESERVED_dtor;
	} else parse_error_pt (p, CODE [p], "Bad basetype");

	if (b->basetype == BASETYPE_SKIP) return p;

	p = collect_dclqual (b, p);
	sumup_dclqual (b);
	p = collect_dclqual (b, p);
	return p;
}

static void ctor_basetype (Dclbase *b, recID r)
{
	if (b->shortname [0] != -1)
		parse_error_ll ("Constructor may not have return type");
	b->shortname [0] = RESERVED_void;
	b->shortname [1] = -1;
	b->basetype = B_VOID;
}

static void dtor_basetype (Dclbase *b, recID r)
{
	if (b->shortname [0] != -1)
		parse_error_ll ("Constructor may not have return type");
	sintprintf (b->shortname, RESERVED_void, '*', -1);
	b->basetype = typeID_voidP;
}

typedef struct {
	Token cname;
	recID rec;
	int dclhype [128];
	Token dclstr [512];
	int dclhi, ip;
	NormPtr objp;
	int argc;
	Token argv [32];
	NormPtr argp [32];
	Token *dflt_arge [32];
	int dflt_argc;
	bool maybe_ctor;
	bool reference;
	bool constant;
	bool const_in_base;
	int ctordtor;
	NormPtr begins;
	bool have_this;
} Dclobj;

/* add 'this' keyword in function prototype */
static void add_this (Dclobj *d, bool const_this)
{
	int i;
	Token *tmp = allocaint (intlen (d->dclstr));

	/* alter the output declaration */
	if (d->dclstr [d->objp + 1] != '(') parse_error_ll ("BUG 234");
	i = d->objp + 2;
	intcpy (tmp, d->dclstr + i);
	if (const_this) d->dclstr [i++] = RESERVED_const;
	d->dclstr [i++] = RESERVED_struct;
	d->dclstr [i++] = name_of_struct (d->rec);
	d->dclstr [i++] = '*';
	d->dclstr [i++] = RESERVED_const;
	d->dclstr [i++] = RESERVED_this;
	if (tmp [0] != ')')
		d->dclstr [i++] = ',';
	intcpy (d->dclstr + i, tmp);
	d->ip = intlen (d->dclstr);
	intcpy (tmp, d->argv);
	d->have_this = true;
}


static typeID dcltype (Dclbase *b, Dclobj *d)
{
	int typet [64];
	if (b->basetype == -1) {
		int *td = open_typeID (b->basetype2);
		typet [0] = td [0];
		intcpy (typet + 1, d->dclhype);
		intcat (typet, td + 1);
	} else {
		typet [0] = b->basetype;
		typet [1] = -1;
		if (d) intcat (typet, d->dclhype);
	}
	if (d->reference)
		typet [0] += REFERENCE_BOOST;
	return enter_type (typet);
}

static NormPtr dirdcl (Dclobj *d, NormPtr p);

static NormPtr dcl (Dclobj *d, NormPtr p)
{
	int ns = 0, ps;

	/* Feature: references */
	if (!d->ip && (d->reference = CODE [p] == '&'))
		++ns, ++p, d->dclstr [d->ip++] = '*';
	/* &-&-&-&-&-&-&-&-&-& */

	for (;; d->dclstr [d->ip++] = CODE [p++])
		if (CODE [p] == '*') ns++;
		else if (CODE [p] != RESERVED_const
			&& CODE [p] != RESERVED_volatile
			&& CODE [p] != RESERVED___restrict) break;
	p = dirdcl (d, ps = p);
	while (ns--) d->dclhype [d->dclhi++] = '*';
	d->constant = CODE [ps - 1] == RESERVED_const && p - ps == 1;
	return p;
}

static NormPtr arglist (Dclobj*, NormPtr);

static int count_argc (NormPtr p)
{
	if (CODE [p] == ')') return 0;
	int c = 1;
	while (CODE [p] != ')')
		switch (CODE [p++]) {
		 case '(': p = skip_parenthesis (p);
		ncase ',': ++c;
		}
	return c;
}

static NormPtr operator_overload (NormPtr p, int memb)
{
	Token op = CODE [p];

	if (CODE [p] == '[') ++p;

	if (!isoperator (op))
		if (isoperfunc (op))
			return p;
		else parse_error_pt (p, op, "not an operator");
	if (CODE [p + 1] != '(')
		parse_error (p, "operator overload is a function");
	memb += count_argc (p + 2);
	CODE [p] = memb == 1 ? isunary_overloadable (op) : isbinary_overloadable (op) ?:
		   isunary_postfix_overloadable (op);
	if (!CODE [p]) parse_error (p, "Can't overload that");
	return p;
}

static NormPtr dirdcl (Dclobj *d, NormPtr p)
{
	if (CODE [p] == '(' && (d->maybe_ctor ? !is_expression (p) : 1)) {
		d->dclstr [d->ip++] = '(';
		p = dcl (d, p + 1);
		d->dclstr [d->ip++] = CODE [p];
		if (CODE [p++] != ')')
			parse_error (p, "Missing parenthesis in declaration");
	} else if (ISSYMBOL (CODE [p])) {
		/* Feature: member functions in global */
		if (CODE [p + 1] == '.') {
			Token bn, n, bm;
			NormPtr ppp = p;
			bn = bm = CODE [p];
			p += 2;
			if (bn == bt_macro) bn = bt_replace;
			/* Feature: ctor with class name */
			if (CODE [p] == bm)
				CODE [p] = RESERVED_ctor;
			else if (CODE [p] == '~' && CODE [p + 1] == bm)
				CODE [++p] = RESERVED_dtor;
			else if (CODE [p] == RESERVED_operator)
				p = operator_overload (p + 1, 1);
			else if (!ISSYMBOL (CODE [p]))
				parse_error (p, "Missing member");
			d->cname = CODE [p];
			d->ctordtor = CODE [p] == RESERVED_ctor ? 1 :
					CODE [p] == RESERVED_dtor ? 2 : 0;
			d->rec = lookup_struct (bn);
			if (!d->rec) {
				/* Feature: out-of-class template func */
				if (!is_template_function (&ppp, d->begins))
					parse_error_pt (p, bn, "No such class");
				d->ctordtor = BASETYPE_SKIP;
				return ppp;
			}
			d->dclstr [d->objp = d->ip++] = 
			n = name_member_function (d->rec, CODE [p]);
			++p;
		} else
		/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
			d->dclstr [d->objp = d->ip++] = CODE [p++];
	} else if (CODE [p] == RESERVED_this)
		d->dclstr [d->objp = d->ip++] = CODE [p++];
	else if (CODE [p] == RESERVED_operator) {
		/* Feature: operator overloading */
		p = operator_overload (p + 1, top_scope > 0);
		d->dclstr [d->objp = d->ip++] = CODE [p++];
		/* *+*+*+*+*+*+*+*+*+*+*+*+*+*+* */
	} else if (CODE [p] == '~' && CODE [p + 1] == RESERVED_dtor) {
		d->dclstr [d->objp = d->ip++] = RESERVED_dtor;
		p += 2;
	} else if (CODE [p] == ESCBASE + '*') {
		/* Kludge: (*) is an escaped operator */
		d->dclstr [d->ip++] = '(';
		d->dclstr [d->ip++] = '*';
		d->dclstr [d->objp = d->ip++] = BLANKT;
		d->dclstr [d->ip++] = ')';
		d->dclhype [d->dclhi++] = '*';
		++p;
	} else d->dclstr [d->objp = d->ip++] = BLANKT;

	for (;;) {
		switch (CODE [p]) {
		case '(':
			/* Constructor: stop at constructor initialization */
			if (!is_dcl_start (CODE [p + 1]) && CODE [p + 1] != ')' && CODE [p + 1] != '.')
				return p;
			/* + + + + + + + + + + + + + + + + + + + + + + + +*/
			d->dclstr [d->ip++] = '(';
			p = arglist (d, p + 1);
			continue;
		case '[':
			d->dclstr [d->ip++] = '[';
			d->dclhype [d->dclhi] = d->dclhi ? '*' : '[';
			d->dclhi++;
			if (CODE [++p] == ']') p += 1;
			else {
				exprret E;
				p = parse_const_expression (p, &E) + 1;
				if (E.ok) {
					NormPtr ps;
					for (ps = 0; E.newExpr [ps] != -1;)
						d->dclstr [d->ip++] = E.newExpr [ps++];
					free (E.newExpr);
				}
			}
			d->dclstr [d->ip++] = ']';
			continue;
		}
		break;
	}
	return p;
}

static NormPtr append_attributes (Dclobj *d, NormPtr p)
{
	if (CODE [p] == RESERVED___attribute__) {
		if (CODE [p + 1] != '(') parse_error (p, "__attribute__ '('");
		NormPtr p2 = skip_parenthesis (p + 2);
		while (p < p2)
			d->dclstr [d->ip++] = CODE [p++];
	}
	return p;
}

static NormPtr parse_dcl (Dclobj *d, NormPtr p)
{
	d->have_this = 0;
	d->argc = -1;
	d->rec = 0, d->objp = -1;
	d->dclhi = d->ip = 0;
	d->cname = -1;
	d->ctordtor = 0;
	d->argv [0] = -1;
	p = dcl (d, p);
	p = append_attributes (d, p);
	if (!d->constant && !d->dclhi)
		d->constant = d->const_in_base;
	d->dclhype [d->dclhi] = -1;
	d->dclstr [d->ip] = -1;
	return p;
}

static NormPtr arglist (Dclobj *d, NormPtr p)
{
	int i, nb = 0;
	typeID t;
	Dclbase b;
	Dclobj di;

	b.class_name = 0;
	di.maybe_ctor = false;
	di.const_in_base = false;
	d->dclhype [d->dclhi++] = '(';

	d->dflt_argc = 0;

	if (CODE [p] == RESERVED_void && CODE [p + 1] == ')') {
		++p;
	} else while (CODE [p] != ')') {
		if (CODE [p] == ELLIPSIS) {
			d->dclstr [d->ip++] = ELLIPSIS;
			d->dclhype [d->dclhi++] = B_ELLIPSIS;
			if (CODE [++p] != ')')
				parse_error (p, "'...' must be last");
			break;
		}
		p = parse_dcl (&di, parse_basetype (&b, p));
		if (b.basetype == BASETYPE_ABSTRACT)
			parse_error (p, "abstract is not a real base type");
		//if (di.dclhype [0] == '[') di.dclhype [0] = '*';
		t = d->dclhype [d->dclhi++] = dcltype (&b, &di);
		for (i = 0; b.dclqual [i] != -1; i++)
			d->dclstr [d->ip++] = b.dclqual [i];
		for (i = 0; b.shortname [i] != -1; i++)
			d->dclstr [d->ip++] = b.shortname [i];
		/* Feature: structure by reference */
		if (isstructure (t) && StructByRef)
			d->dclstr [d->ip++] = '*';
		/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

		if (di.dclstr [di.objp] == BLANKT)
			di.dclstr [di.objp] = internal_identifiern (nb++);

		for (i = 0; di.dclstr [i] != -1; i++)
			d->dclstr [d->ip++] = di.dclstr [i];

		d->argv [++d->argc] = di.dclstr [di.objp];

		/* Feature: default function arguments */
		if (CODE [p] == '=') {
			NormPtr p2;
			p = skip_expression (CODE, p2 = p + 1, INIT_EXPR);
			d->dflt_arge [d->dflt_argc++] = intndup (CODE + p2, p - p2);
		} else if (d->dflt_argc)
			parse_error (p, "Default arguments ought to be last");
		/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

		if (CODE [p] == ',') {
			d->dclstr [d->ip++] = ',';
			p++;
		}
	}
	d->dflt_arge [d->dflt_argc] = 0;
	d->dclstr [d->ip++] = ')';
	d->dclhype [d->dclhi++] = INTERNAL_ARGEND;
	d->argv [++d->argc] = -1;
	return p + 1;
}

static inline NormPtr parse_bitfield (Dclobj *d, NormPtr p)
{
	if (CODE [p] == ':') {
		d->dclstr [d->ip++] = ':';
		exprret E;
		p = parse_const_expression (++p, &E);
		if (E.ok) {
			NormPtr ps;
			for (ps = 0; E.newExpr [ps] != -1;)
				d->dclstr [d->ip++] = E.newExpr [ps++];
			free (E.newExpr);
		}
		d->dclstr [d->ip] = -1;
	}
	return p;
}

bool is_array_of_ctorable_objects (typeID t)
{
	int *tt = open_typeID (t);
	return tt [1] == '[' && tt [2] == -1 && tt [0] >= 0 && need_construction (tt [0]);
}

bool is_array_of_objects (typeID t)
{
	int *tt = open_typeID (t);
	return tt [1] == '[' && tt [2] == -1 && tt [0] >= 0;
}

/* mark the position of a function definition to be parsed later */
static void mark_function_def
	 (Dclbase *b, Dclobj *d, NormPtr p, recID r, typeID rett, Token cname, bool dto)
{
	if (Streams_Closed)
		return;

	Token *tmp = allocaint (intlen (b->dclqual) +
			intlen (b->shortname) + intlen (d->dclstr) + 3);
	intcpy (tmp, b->dclqual);
	intcat (tmp, b->shortname);
	intcat (tmp, d->dclstr);
	store_definition (d->dclstr [d->objp], tmp, d->argv, d->dclhype + 1,
			  p, r, rett, cname, dto, b->isauto ? DT_AUTO : DT_NORM, !b->ismodular);
}

static void add_dclqual (Token q, Token *ts)
{
	int i;
	for (i = 0; ts [i] != -1; i++)
		if (ts [i] == q) return;
	ts [i++] = q;
	ts [i] = -1;
}

static Token pdcl_function (Dclbase *b, Dclobj *d, typeID t, int fflagz)
{
	Token f = d->dclstr [d->objp];
	Token **dflt_arg;
	Token *tmp = allocaint (intlen (b->dclqual) +
			intlen (b->shortname) + intlen (d->dclstr) + 3);
	Token *tmpargv = d->argv;
	intcpy (tmp, b->dclqual);
	intcat (tmp, b->shortname);
	intcat (tmp, d->dclstr);
	if (d->dflt_argc == 0) dflt_arg = 0;
	else {
		int i = d->dflt_argc;
		dflt_arg = (Token**) malloc ((i + 1) * sizeof (Token*));
		memcpy (dflt_arg, d->dflt_arge, i * sizeof (Token*));
		dflt_arg [i] = 0;
	}
	if (d->have_this) {
		tmpargv = allocaint (intlen (d->argv) + 2);
		tmpargv [0] = RESERVED_this;
		intcpy (tmpargv + 1, d->argv);
	}
	return d->rec ?
	 declare_function_member (d->rec, d->cname, t, tmp, tmpargv, fflagz, dflt_arg) :
	 xdeclare_function (&Global, f, f, t, tmp, tmpargv, fflagz, dflt_arg)->name;
}

static NormPtr virtual_class_fwd (recID r, OUTSTREAM D, NormPtr p)
{
	for (;;) {
		if (!ISSYMBOL (CODE [p]))
			parse_error (p, "name expected");
		virtual_inheritance_decl (D, r, CODE [p++]);
		if (CODE [p] == ';') return p + 1;
		if (CODE [p++] != ',')
			parse_error (p, "separator");
	}	
}

/* * * * * * Get the name of a class based on the local typedefs * * * * * */
static void mk_specialized (char nm[], NormPtr p, Token have[], bool anydcl)
{
	char tmp [64];
	int i;
	while (CODE [p] != '}' && CODE [p] != -1) {
		if (CODE [p] == RESERVED_const
		&& ISSYMBOL (CODE [p + 1]) && CODE [p + 3] == ';') {
			p = skip_declaration (p);
			continue;
		}

		if (CODE [p] != RESERVED_typedef)
			if (anydcl) {
				p = skip_declaration (p);
				continue;
			} else parse_error (p, "specialized object may not have"
						" declarations other than typedefs");

		Dclbase b = { .dclqual_i = 0, .class_name = -1 };
		Dclobj d;
		typeID dt;

		p = parse_basetype (&b, last_location = p);
		d.const_in_base = b.isconst;

		if (b.basetype == BASETYPE_ABSTRACT)
			return;

		for (;;) {
			p = parse_dcl (&d, p);

			if (d.dclstr [d.objp] == BLANKT)
				parse_error (p, "Abstract typedef declaration in structure?");

			dt = dcltype (&b, &d);

			for (i = 0; have [i] != -1; i++)
				if (have [i] == d.dclstr [d.objp])
					goto do_not;

			have [i++] = d.dclstr [d.objp];
			have [i] = -1;
			strcat (strcat (strcat (nm, "_"),
				expand (d.dclstr [d.objp])), nametype (tmp, dt));

		do_not:
			if (CODE [p] == ';') break;
			if (CODE [p++] != ',')
				parse_error_pt (p, CODE [p-1], "parse error in declaration separator");
		}
		++p;
	}
}

static void specialize_do (char nm[], Token a, Token have[])
{
	SAVE_VAR (bt_macro, a);
	SAVE_VAR (bt_replace, a);
	Token *ab_par = parents_of_abstract (a);
	mk_specialized (nm, dcl_of_abstract (a), have, true);
	while (*ab_par != -1)
		specialize_do (nm, *ab_par++, have);
	RESTOR_VAR (bt_macro);
	RESTOR_VAR (bt_replace);
}

static Token specialized (Token c, NormPtr p)
{
	char name [1024];
	Token have [64] = { -1, };

	sprintf (name, "%s_specialize_", expand (c));
	mk_specialized (name, p, have, false);
	specialize_do (name, c, have);
	return new_symbol (strdup (name));
}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#define INHERIT_FLAG(x, y) b.x |= flagz & FUNCP_ ## y; if (b.x) flagz |= FUNCP_ ## y

static NormPtr struct_declaration (recID r, OUTSTREAM D, NormPtr p)
{
	NormPtr ps, dstart;
	Dclbase b;
	Dclobj d;
	typeID dt;
	bool isdtor;

	/* Feature: skip local typedefs */
	if (CODE [p] == RESERVED_typedef)
		return skip_declaration (p);
	/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

	b.dclqual_i = 0;
	last_location = p;

	/* Decorative feature: private && public disappear */
	if ((CODE [p] == RESERVED_public || CODE [p] == RESERVED_private)
	&& CODE [p + 1] == ':')
		return p + 2;
	/* ::::::::::::::::::::::::::::::::::::::::::::::: */

	if (CODE [p] == ';') return p + 1;

	/* Feature: force creation of virtual table at object */
	if (CODE [ps = collect_dclqual (&b, p)] == ';') {
		sumup_dclqual (&b);
		if (!b.isvirtual)
			parse_error (p, "bad specifications");
		Here_virtualtable (D, r, b.isinline, b.isconst);
		return ps + 1;
	}
	/* /////////////////////////////////////////////// */

	/* Feature: "virtual class" */
	if (CODE [p] == RESERVED_virtual && CODE [p + 1] == RESERVED_class)
		return virtual_class_fwd (r, D, p + 2);
	/* //////////////////////// */

	/* Feature: class const macros */
	if (CODE [p] == RESERVED_const && ISSYMBOL (CODE [p + 1]) && !is_typename (CODE [p + 1])) {
		enter_class_const (r, CODE [p + 1], CODE [p + 2]);
		if (CODE [p + 3] != ';')
			parse_error (p, "const X Y ';'");
		return p + 4;
	}
	/* ////////////////////////// */

	b.class_name = name_of_struct (r);
	p = parse_basetype (&b, dstart = p);
	d.maybe_ctor = b.maybe_ctor;
	d.const_in_base = b.isconst;

	if (b.basetype == BASETYPE_ABSTRACT)
		if (CODE [p] == ';') return p + 1;
		else parse_error (p, "It is not possible to have objects of abstract classes");
	if (b.basetype == BASETYPE_SKIP)
		return p;

	for (;;) {
		if (CODE [p] == ':') {
			d.ip = 0;
			p = parse_bitfield (&d, p);
			outprintf (D, ISTR (b.dclqual), ISTR (b.shortname),
				   ISTR (d.dclstr), ';',  -1);
			goto skippy;
		}

		p = parse_bitfield (&d, parse_dcl (&d, ps = p));

		if (d.dclstr [d.objp] == BLANKT) {
			/* Feature: anonymous transparent unions */
			if (b.anon_union && d.dclstr [1] == -1) {
				d.dclstr [d.objp] = add_anonymous_union (r, b.basetype);
				goto print_member;
			}
			/* ..................................... */
			parse_error (p, "Abstract declaration in structure?");
		}

		isdtor = false;

		/* Constructors: add return type to ctor function */
		if (d.dclstr [d.objp] == RESERVED_ctor) {
			ctor_basetype (&b, r);
			if (d.dclhype [0] != '(')
				parse_error (p, "Constructor is a function");
		}
		/* +^+^+^+^+^+^+^+^+^+^+^+^+^+^+^+^+^+^+^+^+^+^+ */
		if ((isdtor = d.dclstr [d.objp] == RESERVED_dtor)) {
			dtor_basetype (&b, r);
			if (d.dclhype [0] != '(')
				parse_error (p, "Destructor is a function");
			if (d.dclhype [1] != INTERNAL_ARGEND)
				parse_error (p, "Destructor may not have arguments");
		}
		/* -^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^-^- */

		dt = dcltype (&b, &d);
		if (isfunction (dt)) {
			int flagz = 0;

			/* Feature: class's qualifiers apply to member functions */
			append_dclqual (outer_dclqual, &b);
			/* ------------------------------- */

			bool const_fn = CODE [p] == RESERVED_const;
			if (const_fn) ++p, flagz |= FUNCP_CTHIS;

			bool pure_virtual = CODE [p] == '=' && CODE [p + 1] == RESERVED_0;

			Token fname = d.cname = d.dclstr [d.objp], mfname;

			/* Feature: member functions */
			d.rec = r;
			dt = makemember (dt, r);
			/* Inherited specifiers */
			flagz |= inherited_flagz (r, fname, dt);
			INHERIT_FLAG (isvirtual, VIRTUAL);
			INHERIT_FLAG (isauto, AUTO);
			INHERIT_FLAG (ismodular, MODULAR);
			if (pure_virtual) flagz |= FUNCP_PURE;
			if (!b.ismodular)
				add_this (&d, const_fn);

			/* definitions in-class, add static inline */
			if (CODE [p] == '{') {
				add_dclqual (RESERVED_static, b.dclqual);
				add_dclqual (RESERVED_inline, b.dclqual);
				if (b.isvirtual)
					flagz |= FUNCP_LINKONCE, b.islinkonce = 1;
			}

			mfname = d.dclstr [d.objp] = name_member_function (r, d.dclstr [d.objp]);
			d.dclstr [d.objp] = pdcl_function (&b, &d, dt, flagz);
			if (isdtor) set_dfunc (r, mfname);
			/* ^^^^^^^^^^^^^^^^^^^^^^^^^ */

			/* Feature: virtual nd pure virtual function */
			if (b.isvirtual) {
				if (pure_virtual) p += 2;
				last_location = p;
				Make_virtual_table (D, r, fname, dt);
			}
			/* % % % % % % % % % % % % % % % % % % % % % */

			/* Feature: auto functions */
			if (b.isauto) {
				add_auto_f (d.dclstr [d.objp], fname, r, dt, b.isvirtual,
					    pure_virtual, dstart);
				if (!pure_virtual)
					xmark_section_linkonce (FSP (r), fname, d.dclstr [d.objp]);
			}
			/* ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ */

			 /* Feature: inline members */
			if (CODE [p] == '{') {
				mark_function_def (&b, &d, ++p, r, funcreturn (dt), fname, isdtor);
				if (b.islinkonce)
					xmark_section_linkonce (FSP (r), fname, d.dclstr [d.objp]);
				return skip_braces (p);
			}
			/* ^^^^^^^^^^^^^^^^^^^^^^^^ */

		} else {

			if (isstructure (dt))	/* has virtual table && ok */
				Can_instantiate (base_of (dt));

			/* Feature: Virtual variables */
			if (b.isvirtual || Is_implied_virtual_variable (r, d.dclstr [d.objp])) {
				Token *proto = allocaint (intlen (d.dclstr) + 16);
				sintprintf (proto, ISTR (b.dclqual), ISTR (b.shortname),
					    ISTR (d.dclstr), -1);
				intsubst (proto, d.dclstr [d.objp], MARKER);
				Token *expr = 0;
				if (CODE [p] == '=') {
					exprret E;
					p = parse_const_expression (p + 1, &E);
					if (E.ok) expr = E.newExpr;
				}
				add_virtual_varmemb (r, d.dclstr [d.objp], dt, expr, proto, D);
			/* %%%%%%%%%%%%%%%%%%%%%%%%%% */
			} else {
				add_variable_member (r, d.dclstr [d.objp], dt, d.constant);
			print_member:
				outprintf (D, ISTR (b.dclqual), ISTR (b.shortname),
					   ISTR (d.dclstr), ';',  -1);
			}
		}
	skippy:
		if (CODE [p] == ';') break;
		if (CODE [p++] != ',')
			parse_error_pt (p, CODE [p-1], "parse error in declaration separator");
	}

	return p + 1;
}

/* Feature: parse local typedef member */
static NormPtr struct_typedef (recID r, NormPtr p)
{
	Dclbase b;
	Dclobj d;
	typeID dt;

	if (CODE [p] != RESERVED_typedef)
		return skip_declaration (p);

	b.dclqual_i = 0;
	b.class_name = name_of_struct (r);
	p = parse_basetype (&b, last_location = p);
	d.const_in_base = b.isconst;

	if (b.basetype == BASETYPE_ABSTRACT)
		if (CODE [p] == ';') return p + 1;
		else parse_error (p, "It is not possible to have objects of abstract classes");

	for (;;) {
		p = parse_dcl (&d, p);

		if (d.dclstr [d.objp] == BLANKT)
			parse_error (p, "Abstract typedef declaration in structure?");

		dt = dcltype (&b, &d);

		if (!have_local_typedef (r, d.dclstr [d.objp])) {
			enter_typedef (d.dclstr [d.objp] = add_local_typedef (r, d.dclstr [d.objp]), dt);
			outprintf (GLOBAL, ISTR (b.dclqual), ISTR (b.shortname),
				   ISTR (d.dclstr), ';',  -1);
		}

		if (CODE [p] == ';') break;
		if (CODE [p++] != ',')
			parse_error_pt (p, CODE [p-1], "parse error in declaration separator");
	}

	return p + 1;
}
/* ----------------------------- */

static void parse_all_typedefs (recID r, NormPtr p)
{
	while (CODE [p] != '}' && CODE [p] != -1)
		p = struct_typedef (r, p);
}

static NormPtr parse_all_declarations (recID r, OUTSTREAM o, NormPtr p)
{
static	recID parsing;
	if (parsing)
		set_depend (parsing, r);
	SAVE_VAR (parsing, r);

	while (CODE [p] != '}' && CODE [p] != -1)
		p = struct_declaration (r, o, p);
	if (CODE [p++] == -1)
		parse_error_ll ("missing '}' in structure declaration");

	RESTOR_VAR (parsing);
	return p;
}

static void include_abstract (Token a, recID r, OUTSTREAM o, bool typedefs)
{
	SAVE_VAR (bt_macro, a);

	if (!typedefs) enter_abstract_derrived (a, r);
	Token *ab_par = parents_of_abstract (a);
	if (typedefs) parse_all_typedefs (r, dcl_of_abstract (a));
	else parse_all_declarations (r, o, dcl_of_abstract (a));
	while (*ab_par != -1)
		include_abstract (*ab_par++, r, o, typedefs);

	RESTOR_VAR (bt_macro);
}

static void reparse_auto_functions (recID r, OUTSTREAM o)
{
	current_scope [++top_scope] = r;

	NormPtr rep [256];
	int i;

	borrow_auto_decls (r, rep);

	for (i = 0; rep [i] != -1; i++)
		struct_declaration (r, o, rep [i]);

	--top_scope;
}

static NormPtr parse_structure (Token name, Dclbase *b, NormPtr p, Token s_or_u, Token *ab_par)
{
	recID r = b->basetype, alias;

	current_scope [++top_scope] = r;
	SAVE_VAR (outer_dclqual, b);
	SAVE_VAR (bt_replace, name_of_struct (r));

#ifdef	DEBUG
	if (debugflag.DCL_TRACE)
		PRINTF ("Entering structure ["COLS"%s"COLE"]\n", expand (name_of_struct (r)));
#endif

	OUTSTREAM STRUCTDCL = Streams_Closed ? 0 : new_stream ();
	outprintf (STRUCTDCL, s_or_u, name, '{', -1);

	/* Feature: Inheritance */
	output_parents (STRUCTDCL, r);
	/* ^^^^^^^^^^^^^^^^^^^^ */

	SAVE_VAR (ab_par, ab_par);

	/* Feature: one pass for typedefs  */
	parse_all_typedefs (r, p);
	while (*ab_par != -1)
		include_abstract (*ab_par++, r, STRUCTDCL, 1);
	/* + + + + + + + + + + + + + + + + */

	RESTOR_VAR (ab_par);

	/* Feature: expand abstract parents */
	while (*ab_par != -1)
		include_abstract (*ab_par++, r, STRUCTDCL, 0);
	/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

	p = parse_all_declarations (r, STRUCTDCL, p);

	/* Feature: reparse auto functions from parent decls */
	reparse_auto_functions (r, STRUCTDCL);
	/* ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ */

	--top_scope;
	RESTOR_VAR (outer_dclqual);
	RESTOR_VAR (bt_replace);

	alias = complete_structure (STRUCTDCL, r);
	output_itoken (STRUCTDCL, '}');
	if (CODE [p] == RESERVED___attribute__) {
		NormPtr p2 = p++;
		if (CODE [p++] != '(') parse_error (p, "__attribute__ '('");
		p = skip_parenthesis (p);
		outprintf (STRUCTDCL, NSTRNEXT, CODE + p2, p - p2, - 1);
	}
	output_itoken (STRUCTDCL, ';');

	Token *dcl;
	if (alias) {
		/* Feature: alias classes are just typedefs */
		free_stream (STRUCTDCL);
		outprintf (GLOBAL, RESERVED_typedef,
			    RESERVED_struct, name_of_struct (alias),
			    name_of_struct (r), ';', -1);
		dcl = 0;
		/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
	} else dcl = combine_output (STRUCTDCL);

	set_declaration (r, dcl);
	return p;
}

/* * * Parse the inheritance specifications * * */
/* * * support for abstract classes * * */
static NormPtr parse_inheritance (NormPtr p, recID r, Token aparents[])
{
	bool vp;
	recID parents [32];
	int i = 0, j = 0;

	if ((vp = CODE [p] == RESERVED_virtual))
		p++;
	if (!ISSYMBOL (CODE [p])) parse_error (p, "Expected parent class");

	if (have_abstract (CODE [p]))
		aparents [j++] = CODE [p++];
	else if (!(parents [i++] = lookup_struct (CODE [p++])))
		parse_error_pt (p, CODE [p-1], "No such class");

	if (vp && i) parents [0] += VIRTUALPAR_BOOST;

	while (CODE [p] == ',') {
		++p;
		if ((vp = CODE [p] == RESERVED_virtual))
			++p;
		if (!ISSYMBOL (CODE [p]))
			parse_error_pt (p, CODE [p], "Expected parent class");
	
		if (have_abstract (CODE [p]))
			aparents [j++] = CODE [p++];
		else {
			parents [i] = lookup_struct (CODE [p++]);
			if (!parents [i])
				parse_error_pt (p, CODE [p-1], "No such class");
			if (vp) parents [i] += VIRTUALPAR_BOOST;
			++i;
		}
	}

	if (CODE [p] != '{')
		parse_error (p, "Syntax error after parent classes");

	parents [i] = -1;
	set_parents (r, parents);
	aparents [j] = -1;
	return p;
}
/* -------------------------------------------- */

static NormPtr copy_designator (NormPtr p, Token des [])
{
	int i = 0;
	while (1) switch (CODE [p++]) {
	case '.':
		des [i++] = '.';
		des [i++] = CODE [p++];
		continue;
	case '[': {
		exprret E;
		NormPtr p2 = p;
		des [i++] = '[';
		p = parse_expression (p, &E, NORMAL_EXPR);
		if (CODE [p++] != ']')
			parse_error (p, "Das is verbotten");
		while (p2 < p)
			des [i++] = CODE [p2++];
		}
		continue;
	default: --p;
	case '=':
		des [i++] = '=';
		des [i] = -1;
		return p;
	}
}


static NormPtr aggregate_initializer (OUTSTREAM O, NormPtr p)
{
	exprret E;
	NormPtr p2;

	output_itoken (O, '{');
	while (1) {
		if (CODE [p] == '}') {
			output_itoken (O, '}');
			return p + 1;
		}
		if (CODE [p] == '.' || CODE [p] == '[') {
			Token designator [128];
			p = copy_designator (p, designator);
			outprintf (O, ISTR (designator), -1);
		}
		if (CODE [p] == '{')
			p = aggregate_initializer (O, p + 1);
		else {
			p = parse_expression (p2 = p, &E, INIT_EXPR);
			if (E.ok) {
				outprintf (O, ISTR (E.newExpr), -1);
				free (E.newExpr);
			}
		}
		if (CODE [p] == '}') {
			output_itoken (O, '}');
			return p + 1;
		}
		if (CODE [p++] != ',')
			parse_error (p, "X Verbotten X");
		output_itoken (O, ',');
	}
}

static NormPtr special_initializer (OUTSTREAM O, NormPtr p, typeID t, Token obj)
{
	exprret E;
	NormPtr p2;
	Token designator [256];

	while (1) {
		if (CODE [p] == '}') {
			output_itoken (O, '}');
			return p + 1;
		}
		designator [0] = obj;
		if (CODE [p] == '.' || CODE [p] == '[')
			p = copy_designator (p, designator + 1);
		else parse_error (p, "This object can only be initialized with designators!");

		p = parse_expression (p2 = p, &E, INIT_EXPR);
		if (E.ok) {
			intcat (designator, E.newExpr);
			free (E.newExpr);
			rewrite_designator (t, designator);
			outprintf (O, ISTR (designator), ';', -1);
		}
		if (CODE [p] == '}')
			return p + 1;
		if (CODE [p++] != ',')
			parse_error (p, "X Verbotten X");
	}
}

static NormPtr local_initializer (OUTSTREAM O, NormPtr p, typeID t)
{
	output_itoken (O, '=');
	if (CODE [p] == '{')
		return aggregate_initializer (O, p + 1);
	exprret E;
	p = parse_expression_retconv (p, &E, t, INIT_EXPR);
	if (E.ok) {
		outprintf (O, ISTR (E.newExpr), -1);
		free (E.newExpr);
	}
	return p;
}

/* Generate code for a constructor call. p points to the ctor
   arguments code. if p is -1 then it's the implict constructor
   added automatically.      */
static NormPtr call_ctor (OUTSTREAM O, Token obj, NormPtr p)
{
	Token *ctexpr;

	if (O == GLOBAL_INIT_FUNC) GlobInitUsed = true;

	if (p != -1) {
		Token ctor = CODE [p] == '(' ? RESERVED_ctor : CODE [p++];
		NormPtr ps = p;

		p = skip_parenthesis (p + 1);
		ctexpr = allocaint (p - ps + 5);
		sintprintf (ctexpr, obj, '.', ctor, -1);
		intextract (ctexpr + 3, CODE + ps, p - ps);
	} else {
		ctexpr = allocaint (6);
		sintprintf (ctexpr, obj, '.', RESERVED_ctor, '(', ')', -1);
	}
	if ((ctexpr = rewrite_ctor_expr (ctexpr))) {
		outprintf (O, ISTR (ctexpr), ';', -1);
		free (ctexpr);
	}
	return p;
}

/* Feature: call the ctor for each element of array of objects */
NormPtr gen_array_ctors (NormPtr p, typeID t, Token obj, int *N, Token **M)
{
	OUTSTREAM O = new_stream ();
	NormPtr ps;
	int rest = *N;
	int cnt = 0;
	bool enc, dot, autosize = rest == 0;
	Token *ctexpr, cexpr [128];

	open_local_scope ();
	enter_local_obj (obj, t);
	output_itoken (O, '{');
	while (CODE [p] != '}') {
		dot = CODE [p] == '.';
		p = skip_expression (CODE, ps = p + dot, INIT_EXPR);
		sintprintf (cexpr, obj, '[', new_value_int (cnt++), ']', '.', -1);
		if (!dot) intcatc (cexpr, RESERVED_ctor);
		if (enc = (CODE [ps] != '(' || CODE [p - 1] != ')') && !dot)
			intcatc (cexpr, '(');
		intextract (cexpr + intlen (cexpr), &CODE [ps], p - ps);
		if (enc) intcatc (cexpr, ')');
		if (ctexpr = rewrite_ctor_expr (intdup (cexpr))) {
			outprintf (O, ISTR (ctexpr), ';', -1);
			free (ctexpr);
		}
		if (CODE [p] == ',') ++p;
		else if (CODE [p] != '}') parse_error (p, "Initializer expression separator");
	}

	if (autosize) *N = cnt;
	else if (rest > cnt) {
		Token cc = internal_identifier3 ();

		outprintf (O, RESERVED_int, cc, ';',
			   RESERVED_for, '(', cc, '=', new_value_int (cnt), ';',
			   cc, '<', new_value_int (rest), ';', cc, PLUSPLUS, ')', -1);
		enter_local_obj (cc, typeID_int);
		sintprintf (cexpr, obj, '[', cc, ']', '.', RESERVED_ctor, '(', ')', -1);
		ctexpr = rewrite_ctor_expr (intdup (cexpr));
		outprintf (O, ISTR (ctexpr), ';', -1);
		free (ctexpr);
	} else if (rest < cnt)
		parse_error_ll ("Excess elements in array ctors!");

	output_itoken (O, '}');
	close_local_scope ();
	*M = combine_output (O);

	return p + 1;
}

static NormPtr multiple_array_ctors (OUTSTREAM O, NormPtr p, typeID t, Token obj, Token *n, Token **M)
{
	int cnt;

	if (n [intlen (n) - 2] != '[') {
		if (n [intlen (n) - 3] != '[')
			parse_error_tok (obj, "Can't evaluate complex expression for sizeof :(");
		cnt = eval_int (n [intlen (n) - 2]);
	} else  cnt = 0;

	p = gen_array_ctors (p + 1, t, obj, &cnt, M);

	if (!cnt) {
		backspace_token (O);
		outprintf (O, new_value_int (cnt), ']', -1);
	}
	return p;
}

NormPtr local_declaration (OUTSTREAM O, NormPtr p)
{
	bool hadinit, spec_init=spec_init, isdefinition, mult_init=mult_init;
	Dclbase b;
	Dclobj d;
	typeID dt;
	Token *para_code=para_code;

	if (CODE [p] == ';') return p + 1;

	last_location = p;
	b.class_name = 0;
	p = parse_basetype (&b, p);
	d.maybe_ctor = b.maybe_ctor;
	d.const_in_base = b.isconst;

	if (b.basetype == BASETYPE_ABSTRACT)
		parse_error (p, "It is not possible to have objects of abstract classes");
	if (b.basetype == BASETYPE_SKIP)
		return p;

	if (CODE [p] == ';') {
		/* Feature: abstract declaration for an object with a constructor */
		if (b.basetype >= 0 && has_void_ctor (b.basetype)) {
			Token t = internal_identifier1 ();

			Can_instantiate (b.basetype);
			open_local_scope ();
			enter_local_obj (t, dcltype (&b, 0));
			outprintf (O, '{', ISTR (b.dclqual), ISTR (b.shortname), t, ';', -1);
			if (need_construction (b.basetype))
				alloc_and_init_dcl (O, b.basetype, t, false);
			call_ctor (O, t, -1);
			if (has_dtor (b.basetype))
				outprintf (O, dtor_name (b.basetype),
					   '(', '&', t, ')', ';', -1);
			outprintf (O, '}', -1);
			close_local_scope ();
		}
		/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
		return p + 1;
	}

	for (;; p++) {
		p = parse_dcl (&d, p);

		dt = dcltype (&b, &d);

		if (d.dclstr [d.objp] == BLANKT || d.dclstr [d.objp] == RESERVED__) {
			/* Feature: anoymous obj call ctor thingy */
			if (isstructure (dt) && (CODE [p] == '(' 
			|| CODE [p + 1] == '(' && ISSYMBOL (CODE [p]))) {
				Token t = internal_identifier1 ();
				int basetype = base_of (dt);

				Can_instantiate (basetype);
				open_local_scope ();
				enter_local_obj (t, dt);
				outprintf (O, '{', ISTR (b.dclqual), ISTR (b.shortname),
					   t, ';', -1);
				if (need_construction (basetype))
					alloc_and_init_dcl (O, basetype, t, false);
				p = call_ctor (O, t, p);
				if (has_dtor (basetype))
					outprintf (O, dtor_name (basetype),
						   '(', '&', t, ')', ';', -1);
				outprintf (O, '}', -1);
				close_local_scope ();
				if (CODE [p] != ';')
					parse_error (p, "Anonymous object (constructor) ';'");
				break;
			} else
				parse_error (p, "Abstract declaration in local");
			/* * * * * * * * * * * * * * * * * * * * */
		}

		if (isfunction (dt)) {
			pdcl_function (&b, &d, dt, 0);
		} else {

			outprintf (O, ISTR (b.dclqual), ISTR (b.shortname),
				   ISTR (d.dclstr), -1);

			/* Feature: linkonce statics */
			if (!b.istypedef && b.islinkonce)
				output_itoken (O, linkonce_data_f (d.dclstr [d.objp]));
			/* ^^^^^^^^^^^^^^^^^^^^^^^^^ */

			if ((hadinit = CODE [p] == '=')) {
				spec_init = CODE [p + 1] == '{' &&
					 isstructure (dt) && need_construction (base_of (dt));
				mult_init = d.dclstr [intlen (d.dclstr) - 1] == ']' &&
						is_array_of_objects (dt) && CODE [p + 1] == '{' &&
						has_ctors (base_of (dt)) && CODE [p + 2] != '{';
				if (!spec_init)
					if (!mult_init) p = local_initializer (O, p + 1, dt);
					else p = multiple_array_ctors (O, p + 1, dt, d.dclstr
								[d.objp], d.dclstr, &para_code);
			}
			output_itoken (O, ';');


			/* name is declared after initialization expression
			   therefore int x = x; is ok where the local x is
			   assigned the value of the global x  */
			if (b.istypedef) enter_typedef (d.dclstr [d.objp], dt);
			else enter_local_obj (d.dclstr [d.objp], dt);

			isdefinition = !b.istypedef && !b.isextern;

			if (isstructure (dt) && isdefinition)	/* has virtual table && ok */
				Can_instantiate (base_of (dt));

			/* Feature: Set virtual table */
			if (isdefinition) {
			if (isstructure (dt) && need_construction (base_of (dt)))
				alloc_and_init_dcl (O, base_of (dt), d.dclstr [d.objp], false);
			else if (is_array_of_ctorable_objects (dt))
				alloc_and_init_dcl (O, base_of (dt), d.dclstr [d.objp], true);
			}
			/* ^^^^^^^^^^^^^^^^^^^^^^^^^^ */

			/* Feature: aggregate initializer opened */
			if (hadinit && spec_init)
				p = special_initializer (O, p + 2, dt, d.dclstr [d.objp]);
			/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

			/* Feature: append code for multiple array ctors */
			if (hadinit && mult_init) outprintf (O, ISTR (para_code), -1);
			/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

			/* Feature: construction of objects */
			if (CODE [p] == '(' || CODE [p + 1] == '(' && ISSYMBOL (CODE [p]))
				p = call_ctor (O, d.dclstr [d.objp], p);
			else if (!hadinit && isstructure (dt)
			       && has_void_ctor (base_of (dt)))
					call_ctor (O, d.dclstr [d.objp], -1);
			/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

			/* Feature: mark for auto-destruction */
			if (isstructure (dt) && has_dtor (base_of (dt)))
				add_auto_destruction (d.dclstr [d.objp], base_of (dt));
			/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
		}
		if (CODE [p] == '{') parse_error (0, "Nested functions ?");
		if (CODE [p] == ';') break;
		if (CODE [p] != ',')
			parse_error_pt (p, CODE [p],  "unacceptable declaration, separator;");
	}

	return p + 1;
}

static NormPtr global_initializer (OUTSTREAM OSTR, Token obj, NormPtr p, typeID t)
{
	if (CODE [p] == '{') {
		NormPtr p2 = skip_braces (p + 1);
		outprintf (OSTR, '=', NSTRNEXT, CODE + p, p2 - p, - 1);
		return p2;
	}
	exprret E;
	p = parse_expression_retconv (p, &E, t, INIT_EXPR);
	if (E.ok) {
		if (E.isconstant)
			outprintf (OSTR, '=', ISTR (E.newExpr), - 1);
		else {
			outprintf (GLOBAL_INIT_FUNC, obj, '=', ISTR (E.newExpr), ';', - 1);
			GlobInitUsed = true;
		}
		free (E.newExpr);
	}
	return p;
}

static NormPtr global_declaration (NormPtr p, bool linkonce)
{
	Dclbase b;
	Dclobj d;
	typeID dt;
	bool isdefinition;

	if (CODE [p] == ';') return p + 1;

	d.begins = last_location = p;
	b.class_name = 0;
	p = parse_basetype (&b, p);
	d.maybe_ctor = b.maybe_ctor;
	d.const_in_base = b.isconst;

	if (CODE [p] == ';') return p + 1;

	if (b.basetype == BASETYPE_ABSTRACT)
		parse_error (p, "It is not possible to have objects of abstract classes");
	if (b.basetype == BASETYPE_SKIP)
		return p;

	for (;;) {
		p = parse_dcl (&d, p);

		/* Skip extern abstract func */
		if (d.ctordtor == BASETYPE_SKIP)
			return p;
		/*        -  - - -  -        */

		if (d.dclstr [d.objp] == BLANKT)
			parse_error (p, "Missing object name");

		/* Constructors: add type to ctor function */
		if (d.ctordtor && d.rec)
			(d.ctordtor == 1 ? ctor_basetype : dtor_basetype) (&b, d.rec);
		/* + + + + + + + + + + + + + + + + + + + + */

		dt = dcltype (&b, &d);
		if (isfunction (dt)) {
			bool abstract = false, const_fn = false;
			int flagz = 0;

			/* Feature: adding '*this' */
			if (d.rec) {
				if (const_fn = CODE [p] == RESERVED_const) ++p;
				dt = makemember (dt, d.rec);
				flagz |= exported_flagz (d.rec, d.cname, dt) |
					inherited_flagz (d.rec, d.cname, dt);

				if (flagz & FUNCP_VIRTUAL && flagz & FUNCP_UNDEF)
					parse_error (p, "Declare in the class: virtual!");

				/* Feature: auto function definition ? */
				b.isauto |= flagz & FUNCP_AUTO;
				const_fn |= flagz & FUNCP_CTHIS;
				b.ismodular |= flagz & FUNCP_MODULAR;
				if (!b.ismodular) {
					add_this (&d, const_fn);
					if (is_aliasclass (d.rec))
						remove_struct_from_this (d.dclstr);
				}
			}
			/* ^^^^^^^^^^^^^^^^^^^^^^^ */

			if (!abstract)
				d.dclstr [d.objp] = pdcl_function (&b, &d, dt, flagz);

			if (CODE [p] == '{') {
				mark_function_def (&b, &d, ++p, b.ismodular ? 0 : d.rec,
						   funcreturn (dt), d.cname, d.ctordtor == 2);
				if (b.islinkonce || linkonce)
					xmark_section_linkonce (FSP (d.rec), d.cname,
								d.dclstr [d.objp]);
				return skip_braces (p);
			}

		} else {
			OUTSTREAM OSTR = b.istypedef ? GLOBAL : GVARS;

			if (b.istypedef) {
				if (enter_typedef (d.dclstr [d.objp], dt))
					outprintf (OSTR, ISTR (b.dclqual), ISTR (b.shortname),
						   ISTR (d.dclstr), -1);
				goto skipper;
			} else enter_global_object (d.dclstr [d.objp], dt);

			outprintf (OSTR, ISTR (b.dclqual), ISTR (b.shortname),
				   ISTR (d.dclstr), -1);

			if (!b.istypedef && b.islinkonce)
				output_itoken (OSTR, linkonce_data (d.dclstr [d.objp]));


			isdefinition = !b.istypedef && !b.isextern;

			if (isstructure (dt) && isdefinition)	/* if has virtual table ok */
				Can_instantiate (base_of (dt));

			/* Feature: Set virtual table and/or virtual base class */
			if (isdefinition) {
			if (isstructure (dt) && need_construction (base_of (dt)))
				alloc_and_init_dcl (GLOBAL_INIT_FUNC, base_of (dt),
						 d.dclstr [d.objp], false);
			else if (is_array_of_ctorable_objects (dt))
				alloc_and_init_dcl (GLOBAL_INIT_FUNC, base_of (dt),
						 d.dclstr [d.objp], true);
			}
			/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

			/* here we don't need to parse initialization expression
			   before declaring the object name because it's global
			   and the object name could not possibly refer to
			   an outer scope  */
			if (CODE [p] == '(' || ISSYMBOL (CODE [p]) && CODE [p + 1] == '(')
				p = call_ctor (GLOBAL_INIT_FUNC, d.dclstr [d.objp], p);
			else if (CODE [p] == '=')
				p = global_initializer (OSTR, d.dclstr [d.objp], p + 1, dt);
			else if (isstructure (dt) && has_void_ctor (base_of (dt)))
				call_ctor (GLOBAL_INIT_FUNC, d.dclstr [d.objp], -1);

		skipper:
			output_itoken (OSTR, ';');
		}
		if (CODE [p] == ';') break;
		if (CODE [p++] != ',')
			parse_error_pt (p, CODE [p-1], "invalid declaration separator");
	}

	return p + 1;
}

void reparse_template_func (Token m, recID r, NormPtr p)
{
	SAVE_VAR (bt_macro, m);
	SAVE_VAR (bt_replace, name_of_struct (r));
	current_scope [++top_scope] = r;

	global_declaration (p, 1);

	RESTOR_VAR (bt_macro);
	RESTOR_VAR (bt_replace);
	--top_scope;
}

typeID eval_cast (Token **t)
{
	Dclbase b;
	Dclobj d;

	SAVE_VAR (CODE, *t);

	b.class_name = -1;
	d.maybe_ctor = false;
	d.const_in_base = false;

	parse_dcl (&d, parse_basetype (&b, 0));
	if (b.basetype == BASETYPE_ABSTRACT)
		parse_error_ll ("abstract is not a real base type");
	frealloc (t, intlen (b.shortname) + intlen (d.dclstr) + 1);
	intcpy (*t, b.shortname);
	intcat (*t, d.dclstr);

	RESTOR_VAR (CODE);

	return dcltype (&b, &d);
}

static NormPtr alt_translation_unit (NormPtr p)
{
	if (CODE [p] != RESERVED_C)
		output_itoken (INCLUDE, include_sys_header (CODE [p]));

	if (CODE [++p] != '{')
		parse_error (p, "extern \"string\" '{'");
	++p;

	SAVE_VAR (GLOBAL, 0);
	SAVE_VAR (STRUCTS, 0);
	SAVE_VAR (GVARS, 0);
	SAVE_VAR (StructByRef, false);
	SAVE_VAR (Streams_Closed, true);

	while (CODE [p] != '}' && CODE [p] != -1)
		p = global_declaration (p, 0);

	if (CODE [p++] != '}')
		parse_error (p, "missing '}' in extern header region");

	RESTOR_VAR (GLOBAL);
	RESTOR_VAR (STRUCTS);
	RESTOR_VAR (GVARS);
	RESTOR_VAR (StructByRef);
	RESTOR_VAR (Streams_Closed);

	return p;
}

void translation_unit ()
{
	NormPtr p = 0;

	while (CODE [p] != -1)
		switch (CODE [p]) {
		 case RESERVED__lwc_config_:
			p = lwc_config (p + 1);
		ncase RESERVED_extern:
			if (ISVALUE (CODE [p + 1])
			&& type_of_const (CODE [p + 1]) == typeID_charP) {
				p = alt_translation_unit (p + 1);
				continue;
			}
		default:
			p = global_declaration (p, 0);
		}
}
