#include "global.h"

//******************************************************************************
//		Database of the locations of function definitions
//		&& parsing of all of them later
//		&& re-parsing of auto-functions with checks
//******************************************************************************

typeID return_typeID;

typedef struct funcd_t {
	struct funcd_t *next;
	Token	name, cname;
	recID	object;
	bool	have_this;
	Token	*dclstr;
	Token	*argv;
	typeID	*argt;
	NormPtr	body;
	typeID	ret_type;
	bool	is_destructor;
	bool	polymorph;
	Token	in_abstract;
} funcd;

static struct {
	funcd *first, *last;
} definition [3];

static void _store_definition
 (Token name, Token *dclstr, Token *argv, typeID *argt, NormPtr body,
  recID r, typeID rett, Token cname, bool b, int ff, bool poly, bool have_this)
{
	funcd *f = (funcd*) malloc (sizeof * f);
	f->is_destructor = b;
	f->name = name;
	f->cname = cname;
	f->dclstr = intdup (dclstr);
	f->argv = intdup (argv);
	f->argt = argtdup (argt);
	f->body = body;
	f->object = r;
	f->next = 0;
	f->ret_type = rett;
	f->polymorph = poly;
	f->have_this = have_this;
	f->in_abstract = bt_macro;
	if (definition [ff].last) definition [ff].last->next = f;
	else definition [ff].first = f;
	definition [ff].last = f;
}

void store_definition
 (Token name, Token *dclstr, Token *argv, typeID *argt, NormPtr body,
  recID r, typeID rett, Token cname, bool b, deftype dt, bool ht)
{
	_store_definition (name, dclstr, argv, argt, body, r, rett,
			   cname, b, dt, true, ht);
}

#define CODELESS -1

void store_define_dtor (Token name, recID r)
{
	Token argv [] = { -1 };
	typeID argt [] = { pthis_of_struct (r), INTERNAL_ARGEND };
	Token dclstr [] = {
			RESERVED_static, RESERVED_inline, RESERVED_void,
			'*', name, '(', RESERVED_struct,
			name_of_struct (r), '*', RESERVED_this, ')', -1
		};
	_store_definition (name, dclstr, argv, argt, CODELESS, r,
			   typeID_voidP, -1, true, DT_NORM, true, true);
}

/**********************************************************************
	Rename things on overload
**********************************************************************/
void rename_fdb (Token on, Token nn)
{
	int i;
	funcd *f;

	for (i = DT_NORM; i <= DT_AUTO; i++)
		for (f = definition [i].first; f; f = f->next)
			if (f->name == on) {
				intsubst (f->dclstr, on, nn);
				f->name = nn;
				break;
			}
}

void remove_struct_from_def (Token n)
{
	int i;
	funcd *f;

	for (i = DT_NORM; i <= DT_AUTO; i++)
		for (f = definition [i].first; f; f = f->next)
			if (f->name == n) {
				remove_struct_from_this (f->dclstr);
				break;
			}
}
/**********************************************************************
	--- Usage && comparison
	record usage of important things to be able to
	compare whether two functions are the same
**********************************************************************/
typedef enum { AF_NOTYET, AF_TRUE, AF_FALSE, AF_DEPENDS } AutoBool;
enum { USE_END, USE_TCONST, USE_FCALL, USE_VVAR, USE_MEMB, USE_UPCAST };

typedef struct {
	int status;
	union {
		Token t;
		Token *s;
		recID r;
	} u;
} usaged;

static bool textok = true;
static usaged *udata;
static int nudata, nudata_alloc;

static int add_udata ()
{
#define CHUNK 32
	if (nudata + 1 >= nudata_alloc)
		udata = (usaged*) realloc (udata, sizeof (usaged) * (nudata_alloc += CHUNK));
	return nudata++;
}

void usage_tconst (Token v)
{
	if (textok) {
		int n = add_udata ();
		udata [n].status = USE_TCONST;
		udata [n].u.t = v;
	}
}

void usage_fcall (Token v)
{
	if (textok) {
		int n = add_udata ();
		udata [n].status = USE_FCALL;
		udata [n].u.t = v;
	}
}

void usage_notok ()
{
	textok = false;
}

void usage_vvar (Token *t)
{
	if (!t) usage_notok ();
	else if (textok) {
		int n = add_udata ();
		udata [n].status = USE_VVAR;
		udata [n].u.s = t;
	}
}

void usage_memb (Token v)
{
	if (textok) {
		int n = add_udata ();
		udata [n].status = USE_MEMB;
		udata [n].u.t = v;
	}
}

void usage_upcast (recID r)
{
	if (textok) {
		int n = add_udata ();
		udata [n].status = USE_UPCAST;
		udata [n].u.r = r;
	}
}

static usaged *fix_udata ()
{
	textok = true;
	if (!udata) return 0;

	int n = add_udata ();
	udata [n].status = USE_END;
	usaged *r = realloc (udata, sizeof (usaged) * nudata);
	nudata = nudata_alloc = 0;
	udata = 0;
	return r;
}

static AutoBool is_another_autof (Token);

/* AutoBool means:
	AF_TRUE: redefine
	AF_FALSE: dispatch
	AF_DEPENDS: depends
 */
static int compare_usages (usaged *u1, recID rd, usaged *u2, recID rb, Token deps[])
{
	int i, nd = 0, j;

	if (!u1) return AF_FALSE;

	for (i = 0; u1 [i].status != USE_END; i++)
		switch (u1 [i].status) {
		 case USE_TCONST:
			if (u1 [i].u.t != u2 [i].u.t) return AF_TRUE;
		ncase USE_FCALL:
			if (u1 [i].u.t == u2 [i].u.t) continue;
			switch (is_another_autof (u1 [i].u.t)) {
			case AF_TRUE: return AF_TRUE;
			case AF_FALSE: continue;
			default: for (j = 0; j < nd; j++)
				if (deps [j] == u1 [i].u.t)
					break;
				if (j == nd)
					deps [nd++] = u1 [i].u.t;
			}
		ncase USE_VVAR:
			if (intcmp (u1 [i].u.s, u2 [i].u.s))
				return AF_TRUE;
		ncase USE_MEMB: {
			Token p1 [64], p2 [64], *p;
			lookup_variable_member (rd, u1[i].u.t, p1, true);
			lookup_variable_member (rb, u1[i].u.t, p2 + 2, true);
			p2 [1] = is_ancestor (rd, rb, &p, true) == 2 ? POINTSAT : '.';
			p2 [0] = p [0];
			if (intcmp (p1, p2))
				return AF_TRUE;
			}
		ncase USE_UPCAST: 
			/* if different, will catch fcalls */
			if (u1 [i].u.r != u2 [i].u.r)
				continue;
			{
			Token *p1, p2 [64], *p;
			is_ancestor (rd, u1 [i].u.r, &p1, true);
			if (rb != u1 [i].u.r) {
				is_ancestor (rb, u1 [i].u.r, &p, true);
				intcpy (p2 + 2, p);
				p2 [1] = is_ancestor (rd, rb, &p, true) == 2 ? POINTSAT : '.';
				p2 [0] = p [0];
			} else {
				is_ancestor (rd, rb, &p, true);
				intcpy (p2, p);
			}
			if (intcmp (p1, p2))
				return AF_TRUE;
			}
		}

	deps [nd] = -1;

	return nd ? AF_DEPENDS : AF_FALSE;
}
/**********************************************************************
	--- fwd prototype
**********************************************************************/
static OUTSTREAM do_function (funcd*);
/**********************************************************************
	--- Auto functions, store & instantiate
**********************************************************************/
static struct adf {
	struct adf *next;
	Token dname, fname, pname;
	recID r;
	Token *proto, *argv, *dcltext;
	bool textok;
	usaged *udata;
	int status;
	Token *deps;
	bool have_this;
} *first, *last;

static intnode * autoftree;

void commit_auto_define (Token dname, recID r, Token fname, Token pname)
{
	struct adf *a = (struct adf*) malloc (sizeof * a);
	a->next = 0;
	a->dname = dname;
	a->fname = fname;
	a->pname = pname;
	a->r = r;
	a->status = AF_NOTYET;
	a->udata = 0;
	a->deps = 0;
	a->dcltext = a->proto = a->argv = 0;

	if (last) last->next = a;
	else first = a;
	last = a;
	union ival u = { .p = a };
	intadd (&autoftree, a->fname, u);
}

static AutoBool is_another_autof (Token t)
{
	intnode *n = intfind (autoftree, t);
	if (!n) return AF_TRUE;
	struct adf *a = (struct adf*) n->v.p;
	return a->status == AF_NOTYET ? AF_DEPENDS : a->status;
}

static AutoBool check_dependancies (Token dep[])
{
	int i, r = AF_FALSE;

	for (i = 0; dep [i] != -1; i++)
	switch (((struct adf*) intfind (autoftree, dep [i])->v.p)->status) {
	case AF_TRUE: return AF_TRUE;
	case AF_DEPENDS: r = AF_DEPENDS;
	}

	return r;
}

/* instantiate the definition of an auto function for a specific class
 * and then parse it */
static Token *instantiate_auto (struct adf *a)
{
	funcd *f;
	Token *dclstr;

	for (f = definition [DT_AUTO].first; f; f = f->next)
		if (f->name == a->dname)
			break;

	if (!f) return 0;

	a->have_this = f->have_this;
	dclstr = intdup1 (f->dclstr);

	if (a->fname != a->dname) {
		recID r1 = is_aliasclass (a->r);
		recID r2 = is_aliasclass (f->object);

		intsubst1 (dclstr, a->dname, a->fname);
		Token *tt;
		for (tt = dclstr; *tt != a->fname; ++tt);
		intsubst1 (tt, name_of_struct (f->object), name_of_struct (a->r));

		if (r1 && !r2) remove_struct_from_this (dclstr);
		else if (r2 && !r1) add_struct_to_this (dclstr);
	}

	a->proto = dclstr;
	a->argv = intdup (f->argv);

	funcd F = {
		.name = a->fname, .cname = f->cname, .object = a->r,
		.dclstr = dclstr, .argv = f->argv, .argt = f->argt,
		.body = f->body, .ret_type = f->ret_type,
		.is_destructor = f->is_destructor, .polymorph = false,
		.have_this = f->have_this
	};

	return combine_output (do_function (&F));
}

static AutoBool same_code (struct adf *f)
{
	struct adf *a = (struct adf*) intfind (autoftree, f->pname)->v.p;
	Token depends [64];

	if (!a->textok)
		return AF_TRUE;

	AutoBool r = compare_usages (f->udata, f->r, a->udata, a->r, depends);
	if (r == AF_DEPENDS)
		f->deps = intdup (depends);
	return r;
}

static void dispatch_to_parent (struct adf *a)
{
	int i = 0;
	recID r = ((struct adf*) intfind (autoftree, a->pname)->v.p)->r;

	outprintf (FUNCDEFCODE, ISTR (a->proto), '{', RESERVED_return,
		   a->pname, '(', -1);

	if (a->have_this)
 		outprintf (FUNCDEFCODE, ISTR (upcast1_this (a->r, r)), -1);
	else if (a->argv [0] != -1)
		outprintf (FUNCDEFCODE, a->argv [i++], -1);

	for (; a->argv [i] != -1; i++)
		outprintf (FUNCDEFCODE, ',', a->argv [i], -1);
	outprintf (FUNCDEFCODE, ')', ';', '}', -1);
}

void define_auto_functions ()
{
	struct adf *a, *fd;

	//
	// first, parse all of them, hypothetically
	//
	objective.recording = true;
	for (a = first; a; a = a->next)
		if (a->dcltext = instantiate_auto (a))
			a->textok = textok, a->udata = fix_udata ();
	objective.recording = false;

	if (ExpandAllAutos) {
		for (a = first; a; a = a->next)
			if (a->dcltext)
				outprintf (FUNCDEFCODE, ISTR (a->dcltext), -1);
		goto free_Skylarov;
	}

	//
	// now define or dispatch depending whether
	// the two versions do different things
	//
	for (a = first; a; a = a->next)
	if (a->dcltext)
		switch (a->status = (a->pname == -1 ? AF_TRUE : same_code (a))) {
		 case AF_TRUE:
			outprintf (FUNCDEFCODE, ISTR (a->dcltext), -1);
		ncase AF_FALSE:
			dispatch_to_parent (a);
		}

	//
	// autofunctions calling autofunctions
	// try to resolve known
	//
	for (fd = 0, a = first; a; a = a->next)
	if (a->status == AF_DEPENDS) {
		switch (a->status = check_dependancies (a->deps)) {
		 case AF_TRUE:
			outprintf (FUNCDEFCODE, ISTR (a->dcltext), -1);
		ncase AF_FALSE:
			dispatch_to_parent (a);
		ncase AF_DEPENDS:
			if (!fd) fd = a;
			continue;
		}

		if (fd) {
			a = fd;
			fd = 0;
		}
	}

	//
	// circular dependancies left. suppose false
	//
	for (a = first; a; a = a->next)
		if (a->status == AF_DEPENDS)
			dispatch_to_parent (a);

free_Skylarov:
	// (Do you remember malloc?)
	for (a = first; a; a = a->next) {
		if (a->dcltext) free (a->dcltext);
		if (a->udata) free (a->udata);
		if (a->deps) free (a->deps);
		if (a->proto) free (a->proto);
		if (a->argv) free (a->argv);
	}
}
/**********************************************************************
	--- Abstract Class functions, store & instantiate
**********************************************************************/
/**********************************************************************
	Parse all functions
**********************************************************************/
static OUTSTREAM do_function (funcd *f)
{
	int i;
	bool MAIN;	// and gcc warnings get sillier all the time....
	OUTSTREAM o = new_stream ();
#ifdef	DEBUG
	if (debugflag.FUNCPROGRESS)
		PRINTF ("FUNCTION ["COLS"%s"COLE"]\n", expand(f->name));
#endif

	return_typeID = f->ret_type;
	if (MAIN = f->name == RESERVED_main)
		MainModule = true;
	objective.isdtor = f->is_destructor;

	/* abstract template class type */
	SAVE_VAR (bt_macro, f->in_abstract);
	SAVE_VAR (bt_replace, name_of_struct (f->object));
	SAVE_VAR (in_function, f->name);
	current_scope [++top_scope] = f->object;

	/* set expression semantics in member functions */
	if ((objective.yes = f->object != 0)) {
		objective.func = f->name;
		objective.class = f->object;
		objective.efunc = f->cname;
		objective.classP = pthis_of_struct (f->object);
		objective.have_this = f->have_this;
		objective.polymorph = f->polymorph;
	}
	open_local_scope ();
	for (i = 0; f->argt [i] != INTERNAL_ARGEND; i++)
		if (f->argv [i] != -1)
			enter_local_obj (f->argv [i], f->argt [i]);
		else break;
	o = new_stream ();

	outprintf (o, ISTR (f->dclstr), - 1);
	if (f->body != CODELESS)
		compound_statement (o, f->body);
	else outprintf (o, '{', '}', -1);

	/* Feature: dtors call dtors of dtorable members */
	if (f->is_destructor) {
		backspace_token (o);
		if (has_dtorable (f->object))
			produce_dtorables (o, f->object);
		outprintf (o, RESERVED_return, RESERVED_this, ';', '}', -1);
		if (objective.recording)
			usage_notok ();
	}
	/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

	/* Everybody's favorite addition */
	if (MAIN) {
		backspace_token (o);
		outprintf (o, RESERVED_return, RESERVED_0, ';', '}', -1);
	}

	close_local_scope ();

	RESTOR_VAR (bt_macro);
	RESTOR_VAR (bt_replace);
	RESTOR_VAR (in_function);
	--top_scope;

	return o;
}

void do_functions ()
{
	funcd *f;

	FUNCDEFCODE = new_stream ();
#ifdef	COMMENT_OUTPUT
	output_itoken (FUNCDEFCODE, new_symbol (COMMENT_SECTION_FUNCTIONS));
#endif

	for (f = definition [DT_NORM].first; f; f = f->next)
		concate_streams (FUNCDEFCODE, do_function (f));
}
