/*
 *
 * Copyright (C) 2003, Stelios Xanthakis
 *
 */
#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>
#ifdef SYS_HAS_MMAP
#include <sys/mman.h>
#endif
#include <stdarg.h>
#include <setjmp.h>
#include <time.h>


#define COLS "\033[01;37m"
#define COLE "\033[0m"

//*****************************************************************************
// compiler.h
//*****************************************************************************

#ifdef __GNUC__
#define COMPILER "gcc"
#define INTERN_memcpy RESERVED___builtin_memcpy
#define INTERN_alloca RESERVED___builtin_alloca
#else
#define COMPILER "generic"
#define INTERN_memcpy RESERVED_memcpy
#define INTERN_alloca RESERVED_alloca
#endif

//*****************************************************************************
// ld.h
//*****************************************************************************

#define SECTION_LINKONCE_DATA   ".gnu.linkonce.d."
#define SECTION_LINKONCE_TEXT   ".gnu.linkonce.t."
#define SECTION_LINKONCE_RODATA ".gnu.linkonce.r."
#define HAVE_LINKONCE
#define HAVE_BUILTIN_MEMCPY
#define SIZEOF_UCONTEXT "512"	// should suffice for the near future

//*****************************************************************************
// version.h
//*****************************************************************************

#define VERSION_MAJ 1
#define VERSION_MIN 0
#define LWC_VERSION "1.0"

#include "norm.h"

extern int zinit [];

//*****************************************************************************
// global.h
//*****************************************************************************

typedef int NormPtr, Token, recID, typeID;
typedef int bool;

#define true 1
#define false 0

#define ncase break; case
#define ndefault break; default

#ifdef __GCC__
#define alloca __builtin_alloca
#endif

#define allocaint(x) (int*) alloca ((x) * sizeof (int))
#define mallocint(x) (int*) malloc ((x) * sizeof (int))
static inline int* frealloc (int **a, int x) { free (*a); return *a = mallocint (x); }

//*****************************************************************************
// debug.h
//*****************************************************************************

#define DEBUG

#ifdef DEBUG
extern struct debugflag_s {
	bool
	GENERAL,
	PEXPR,
	PEXPR_EXTREME,
	SHOWPROG,
	FUNCPROGRESS,
	OUTPUT_INDENTED,
	PARSE_ERRORS_SEGFAULT,
	EXPR_ERRORS_FATAL,
	DCL_TRACE,
	TDEF_TRACE,
	VIRTUALTABLES,
	VIRTUALBASE,
	VIRTUALCOMBINE,
	AUTOF,
	NOSTDOUT,
	SHOWCPP,
	CPP,
	REGEXP_DEBUG;
} debugflag;

#define PREPROCOUT "preproc.i"

extern void	enable_debugs	();
extern FILE	*logstream;
#define PRINTF(...)	fprintf (logstream, __VA_ARGS__)
#define INTPRINT(x)	fintprint (logstream, x)
#define INTPRINTN(x)	{ INTPRINT(x); PRINTF("\n"); }
#define INTPRINTF(x,y)	{ PRINTF(x); INTPRINT(y); PRINTF("\n"); }
#endif

//*****************************************************************************
// config.h
//*****************************************************************************

#define MAGIC_DIGIT ""
#define GLOBINITF "GLoBaL_ConStRuCtOr"
#define DERRIVE_UNION "un"
#define PREPROCFILE ".preprocfile"
#define COMMENT_OUTPUT

#ifdef COLOR_IN_COMMENTS
#define REM_COLS COLS
#define REM_COLE COLE
#else
#define REM_COLS ""
#define REM_COLE ""
#endif

#define COMMENT_SECTION_FUNCTIONS \
"\n/*"REM_COLS"******** Program function definitions ********"REM_COLE"*/\n"

#define COMMENT_SECTION_OBJCTOR \
"\n/*"REM_COLS"*********** Module _Init Function  ***********"REM_COLE"*/\n"

#define COMMENT_SECTION_PROTOTYPES \
"\n/*"REM_COLS"************ Function Prototypes  ***********"REM_COLE"*/\n"

#define COMMENT_SECTION_INNERCODE \
"\n/*"REM_COLS"************ Internal Functions  ************"REM_COLE"*/\n"

#define COMMENT_SECTION_VIRTUALTABLES \
"\n/*"REM_COLS"*************** Virtual tables ***************"REM_COLE"*/\n"

#define COMMENT_SECTION_VTDEF \
"\n/*"REM_COLS"************* Virtua Table declarations *************"REM_COLE"*/\n"

#define COMMENT_SECTION_STRUCTS \
"\n/*"REM_COLS"************* Structures *************"REM_COLE"*/\n"

#define COMMENT_SECTION_GVARS \
"\n/*"REM_COLS"************* Global variables *************"REM_COLE"*/\n"

#define COMMENT_SECTION_INCLUDE \
"\n/*"REM_COLS"************* system headers *************"REM_COLE"*/\n"

#define COMMENT_SECTION_GLOBAL \
"\n/*"REM_COLS"************* global scope *************"REM_COLE"*/\n"

#define COMMENT_SECTION_REGEXP \
"\n/*"REM_COLS"************* regular expressions *************"REM_COLE"*/\n"

//*****************************************************************************
// preproc.h
//*****************************************************************************

extern void	preproc (int, char**);
extern char	*current_file, *main_file;

//*****************************************************************************
// lwc_config.h
//*****************************************************************************

extern NormPtr	lwc_config (NormPtr);

//*****************************************************************************
// main.h
//*****************************************************************************

extern bool MainModule, HadErrors, Reentrant, ExpandAllAutos;
extern bool StructByRef, InlineAllVt, ConstVtables, VIDeclarations;
extern bool Streams_Closed, ExceptionsUsed, GlobInitUsed, vtptrConst;
extern int max_symbol_len;

//*****************************************************************************
// output.h
//*****************************************************************************

#define EOST		-1
#define STRNEXT		-2
#define NSTRNEXT	-3
#define BLANKT		-4
#define NONULL		-5

#define ISTR(x) STRNEXT, x

typedef struct outstream *OUTSTREAM;

extern OUTSTREAM GLOBAL, GLOBAL_INIT_FUNC, FUNCDEFCODE, INTERNAL_CODE, FPROTOS;
extern OUTSTREAM VTABLE_DECLARATIONS, VIRTUALTABLES, STRUCTS, GVARS, INCLUDE, REGEXP_CODE;

extern OUTSTREAM	new_stream ();

extern void		free_stream	(OUTSTREAM);
extern int		ntokens		(OUTSTREAM);
extern inline void	output_itoken	(OUTSTREAM, int);
extern void		outprintf	(OUTSTREAM, Token, ...);
extern void		backspace_token	(OUTSTREAM);
extern int		*combine_output	(OUTSTREAM);
extern OUTSTREAM	concate_streams (OUTSTREAM, OUTSTREAM);
extern void		export_output	(OUTSTREAM);

//*****************************************************************************
// cpp.h
//*****************************************************************************

extern bool	sys_cpp;
extern void	setup_cpp	(int, char**);
extern void	cleanup_cpp	();
extern void	cpp_directive	();

#define VARBOOST 1000
#define VOIDARG	  999
extern int	is_macro	(Token);
extern Token*	expand_macro	(Token);

//*****************************************************************************
// lex.h
//*****************************************************************************

extern void	fatal		(char*);
extern int	*CODE, c_ntok;
extern int	c_nsym, c_nval;

extern char	*Cpp;
extern int	Ci, Clen;
extern int	line;
extern char	*tfile;

extern Token	enter_symbol	(char*);

extern int	c_line_of	(NormPtr);
extern char	*c_file_of	(NormPtr);

extern void	adjust_lines	(NormPtr, int);

extern void	initlex		();
extern int	yydo		(char*);
extern int	yydo_file	(char*);
extern void 	yydo_mem	(char*, int);
extern char	*expand		(Token);
inline
static char	*EXPC		(NormPtr p) { return expand (CODE [p]); }
extern Token	new_symbol	(char*);
extern Token	new_value_int	(int);

extern void	add_extra_values	(char**, int);

extern Token	Lookup_Symbol	(char*);

extern typeID	type_of_const	(Token);

extern int		eval_int	(Token);
extern long long int	eval_intll	(Token);

extern Token 	binshift [];

#define tokcmp(tok, str) strcmp (expand (tok), str)

//*****************************************************************************
// templates.h
//*****************************************************************************

extern void	do_templates ();

//*****************************************************************************
// util.h
//*****************************************************************************

extern NormPtr	skip_parenthesis	(NormPtr);
extern NormPtr	skip_brackets		(NormPtr);
extern NormPtr	skip_braces		(NormPtr);
extern NormPtr	skip_declaration	(NormPtr);
extern NormPtr	skip_buffer_parenthesis	(Token*, NormPtr);
extern NormPtr	skip_buffer_brackets	(Token*, NormPtr);
extern NormPtr	skip_buffer_braces	(Token*, NormPtr);

extern bool	intchr		(int*, int);
extern void	intsubst	(int*, int, int);
extern void	intsubst1	(int*, int, int);
extern void	intcpy		(int*, int*);
extern void	intcat		(int*, int*);
extern void	intncat		(int*, int*, int);
extern void	intcatc		(int*, int);
extern void	intextract	(int*, int*, int);
extern int	intlen		(int*);
extern int	*intdup		(int*);
extern int	*intdup1	(int*);
extern int	*argtdup	(typeID*);
extern int	*intndup	(int*, int);
extern int	intcmp		(int*, int*);
extern void	intprint	(Token*);
extern void	dintprint	(Token*);
extern void	fintprint	(FILE*, Token*);
extern Token	*sintprintf	(Token*, Token, ...);

extern char*	escape_c_string	(char*, int);
extern char*	escape_q_string	(char*, int);
extern char*	loadtext	(char*);

struct load_file {
    int fd;
    char *data;
    int len;
    int success;
};
extern void ctor_load_file_ (struct load_file *const, char*);
extern void dtor_load_file_ (struct load_file *const);

typedef struct intnode_t {
	struct intnode_t *less, *more;
	int key;
	union ival {
		void *p;
		int i;
	} v;
} intnode;

extern intnode	*intfind	(intnode*, int);
extern void	intadd		(intnode**, int, union ival);
extern void	intremove	(intnode**, intnode*);

extern void	debug_pr_type	(typeID);

//*****************************************************************************
// misc.h
//*****************************************************************************

typedef struct {
	jmp_buf env;
} jmpbufs;

extern void	set_catch		(jmpbufs, NormPtr, Token*);
extern void	expr_error		(char*);
extern void	expr_errort		(char*, Token);
extern void	expr_errortt		(char*, Token, Token);
extern void	expr_error_undef	(Token, int);
extern void	clear_catch		();

extern NormPtr	last_location;
extern Token	in_function;

extern void	parse_error		(NormPtr, char*);
extern void	parse_error_tok		(Token, char*);
extern void	parse_error_cpp		(char*);
extern void	parse_error_toktok	(Token, Token, char*);
extern void	parse_error_pt		(NormPtr, Token, char*);
extern void	parse_error_ll		(char*);

extern typeID	bt_promotion	(typeID);
extern typeID	ptrup		(typeID);
extern typeID	ptrdown		(typeID);
extern typeID	funcreturn	(typeID);
extern typeID	dereference	(typeID);
extern typeID	makemember	(typeID, recID);

extern Token	isunary_overloadable		(Token);
extern Token	isunary_postfix_overloadable	(Token);
extern Token	isbinary_overloadable		(Token);

extern void	remove_struct_from_this	(Token*);
extern void	add_struct_to_this	(Token*);

extern Token	include_sys_header	(Token);

extern Token	linkonce_data	(Token);
extern Token	linkonce_data_f	(Token);
extern Token	linkonce_text	(Token);
extern Token	linkonce_rodata	(Token);

extern bool	is_expression	(NormPtr);

extern void	bogus1	();

//*****************************************************************************
// fspace.h
//*****************************************************************************

typedef struct funcp_t {
	struct funcp_t	*next;
	Token		name;
	typeID		type;
	Token		*prototype;
	typeID		*ovarglist;
	Token		**dflt_args;
	Token		*xargs;
	int		flagz;
	bool		used;
} funcp;

#define FUNCP_LINKONCE	1
#define FUNCP_AUTO	2
#define FUNCP_MODULAR	4
#define FUNCP_CTHIS	8
#define FUNCP_VIRTUAL	16
#define FUNCP_PURE	32
#define FUNCP_UNDEF	64

typedef intnode *fspace;

extern fspace Global;

extern typeID*	promoted_arglist	(typeID*);
extern typeID*	promoted_arglist_t	(typeID);
extern char*	nametype		(char *ret, typeID);
extern char*	type_string		(char*, typeID);
extern bool	arglist_compare		(typeID*, typeID*);
extern funcp*	xdeclare_function	(fspace*, Token, Token, typeID, Token*, Token*,int,Token**);
extern Token	declare_function_member	(recID, Token, typeID, Token*, Token*, int, Token**);
extern funcp*	xlookup_function_dcl	(fspace, Token, typeID[]);

typedef struct {
	Token	oname;
	typeID	t;
	Token	**dflt_args;
	Token	*prototype, *xargs;
	int	flagz;
} flookup;

extern bool	have_function		(fspace, Token);
extern int	xlookup_function	(fspace, Token, typeID[], flookup*);
extern typeID	lookup_function_symbol	(Token);
extern void	xmark_section_linkonce	(fspace, Token, Token);
extern void	export_fspace		(fspace);

//*****************************************************************************
// cdb.h
//*****************************************************************************

extern void	enter_enumconst	(Token);
extern int	is_enumconst	(Token);
extern void	enter_enum	(Token);
extern int	is_enum		(Token);
extern bool	enter_typedef	(Token, typeID);
extern typeID	lookup_typedef	(Token);

extern typeID	enter_type	(int*);
extern int	*open_typeID	(typeID);
extern bool	isfunction	(typeID);
extern int	is_typename	(Token);
extern bool	is_dcl_start	(Token);

extern void	enter_abstract		(Token, Token*, NormPtr);
extern bool	have_abstract		(Token);
extern bool	abstract_has_special	(Token, recID);
extern NormPtr	dcl_of_abstract		(Token);
extern Token*	parents_of_abstract	(Token);

extern bool	is_template_function	(NormPtr*, NormPtr);
extern bool	is_extern_templ_func	(NormPtr*);
extern void	enter_abstract_derrived	(Token, recID);
extern void	specialize_abstracts	();

extern void	enter_global_object	(Token, typeID);
extern typeID	lookup_global_object	(Token);

extern void	open_local_scope	();
extern void	add_catchpoint		(int);
extern void	enter_local_obj		(Token, typeID);
extern void	add_auto_destruction	(Token, recID);
extern bool	scope_has_dtors		();
extern bool	break_has_dtors		(int);
extern bool	func_has_dtors		();
#define		REFERENCE_BOOST		10000
extern typeID	lookup_local_obj	(Token);
extern void	gen_auto_destruction	(OUTSTREAM);
extern void	gen_break_destructors	(OUTSTREAM, int);
extern void	gen_all_destructors	(OUTSTREAM);
extern void	close_local_scope	();
extern void	rmv_catchpoint		();

static inline int base_of (typeID t)
{ return open_typeID (t)[0]; }

enum {
	B_SCHAR = -32, B_UCHAR, B_SSINT, B_USINT, B_SINT, B_UINT,
	B_SLONG, B_ULONG, B_FLOAT, B_DOUBLE, B_LDOUBLE,
	B_VOID, B_ELLIPSIS, INTERNAL_ARGEND /* -19 */
};

#define REFERENCE_BASE (REFERENCE_BOOST + B_SCHAR)

static inline int dbase_of (typeID t)
{ recID r = base_of (t); return r >= REFERENCE_BASE ? r - REFERENCE_BOOST : r; }

extern typeID typeID_int, typeID_float, typeID_charP;
extern typeID typeID_voidP ,typeID_void, typeID_uint;

//*****************************************************************************
// hier.h
//*****************************************************************************

// - build db

#define		VIRTUALPAR_BOOST	10000
extern recID	enter_struct		(Token, bool);
extern void	set_depend		(recID, recID);
extern void	set_declaration		(recID, Token*);
extern void	set_parents		(recID, recID[]);
extern void	output_parents		(OUTSTREAM, recID);
extern void	add_variable_member	(recID, Token, typeID, bool);
extern Token	add_anonymous_union	(recID, recID);
extern Token	add_local_typedef	(recID, Token);
extern Token	have_local_typedef	(recID, Token);
extern void	add_auto_f		(Token, Token, recID, typeID, bool, bool, NormPtr);
extern void	set_dfunc		(recID, Token);
extern Token	dtor_name		(recID);
extern void	enter_class_const	(recID, Token, Token);
extern Token	add_function_member	(recID, Token);
extern int	Here_virtualtable	(OUTSTREAM, recID, bool, bool);
extern void	add_virtual_varmemb	(recID, Token, typeID, Token*, Token*, OUTSTREAM);
extern void	rename_hier		(Token, Token);
extern int	virtual_inheritance_decl	(OUTSTREAM, recID, Token);
typedef struct {
	recID r;
	Token fname;
	typeID ftype;
	Token fmemb;
	Token *prototype, *argv;
	int flagz;
	OUTSTREAM scopeout;
} vf_args;
extern void	Make_virtual_table	(OUTSTREAM, recID, Token, typeID);
extern void	Enter_virtual_function	(vf_args*);
extern recID	complete_structure	(OUTSTREAM, recID);
extern recID	is_aliasclass		(recID);
extern recID	aliasclass		(recID);
#define iRESERVED_struct(r) (is_aliasclass (r) ? BLANKT : RESERVED_struct)

// - use db

extern recID	lookup_struct	(Token);
extern int	is_struct	(Token);
extern Token	name_of_struct	(recID);
static 
inline char*	SNM		(recID r)	{ return expand (name_of_struct (r)); }
extern typeID	pthis_of_struct	(recID);
extern fspace	FSP		(recID);
extern bool	has_const_members	(recID);

extern int	inherited_flagz			(recID, Token, typeID);
extern int	exported_flagz			(recID, Token, typeID);
extern typeID	lookup_variable_member		(recID, Token, Token*, bool);
extern typeID	lookup_virtual_varmemb		(recID, Token, Token*, bool, Token**);
extern Token	lookup_class_const		(recID, Token);
extern int	lookup_function_member		(recID, Token, typeID[], flookup*, bool);
extern int	lookup_virtual_function_member	(recID, Token, typeID[], Token*, flookup*);
extern bool	Is_implied_virtual_variable	(recID, Token);
extern bool	Is_pure_virtual			(recID, Token, typeID[], flookup*);
typedef struct {
	typeID t;
	Token rec, *memb;
} vtvar;
extern vtvar	access_virtual_variable		(recID, Token);

extern bool	has_void_ctor	(recID);
extern bool	has_dtor	(recID);
extern bool	has_dtorable	(recID);
extern bool	has_ctors	(recID);
extern Token	have_fmemb_dcl	(recID, Token);

extern bool	zero_offset		(recID, recID);
extern bool	isancestor		(recID, recID);
extern int	is_ancestor		(recID, recID, Token**, bool);
extern int	is_ancestor_runtime	(recID, recID, Token**);
extern Token*	upcast1_this		(recID, recID);

extern recID	ancest_named	(recID, Token);
extern void	downcast_rtti	(recID, recID, recID, Token[]);

extern void export_virtual_table_instances	();
extern void export_virtual_definitions		();
extern void export_virtual_static_definitions	();
extern void export_structs			();

extern bool Can_instantiate		(recID);
extern bool need_construction		(recID);
extern bool need_vbase_alloc		(recID);

extern void gen_vt_init			(OUTSTREAM, recID, Token, Token);
extern void produce_dtorables		(OUTSTREAM, recID);
extern void gen_construction_code	(OUTSTREAM, recID, Token);

extern Token	lookup_local_typedef	(recID, Token);
extern int	borrow_auto_decls	(recID, NormPtr[]);

//*****************************************************************************
// inames.h
//*****************************************************************************

extern Token	name_anonymous_struct	();
extern Token	name_anonymous_union	();
extern Token	name_anonymous_enum	();
extern Token	name_anon_union		(int, Token);
extern Token	name_overload_fun	(Token, char*);
extern Token	name_member_function	(recID, Token);
extern Token	name_member_function_virtual	(Token, Token);
extern Token	name_local_typedef	(Token, Token);
extern Token	name_internal_object	();
extern Token	internal_identifier1	();
extern Token	internal_identifier2	();
extern Token	internal_identifier3	();
extern Token	internal_identifiern	(int);
extern Token	name_define_new		(Token, Token);
extern Token	name_inherited		(Token);
extern Token	name_storage_inherit	(recID);
extern Token	name_downcast		(Token, Token);
extern Token	name_downcast_safe	(Token, Token);
extern Token	name_upcast_safe	(Token, Token);
extern Token	name_virtual_slot	(recID, Token, typeID);
extern Token	name_virtual_variable	(recID, Token);
extern Token	name_instance		(recID, recID);
extern Token	name_virtual_table	(recID);
extern Token	name_virtual_inner	(recID, recID, Token, typeID);
extern Token	name_name_enumerate	(Token, int);
extern Token	name_rtti_slot		(recID, recID);
extern Token	name_intern_ctor	(recID);
extern Token	name_derrive_memb	(recID);
extern Token	name_anon_regexp	();
extern Token	name_derrive_union;
extern Token	tokstrcat		(Token, char*);

//*****************************************************************************
// icode.h
//*****************************************************************************

extern Token	i_call_initialization	(recID);
extern Token	i_downcast_function	(recID, recID);
extern Token	i_downcast_null_safe	(recID, recID);
extern Token	i_upcast_null_safe	(recID, recID, Token*, bool);
extern Token	i_member_virtual	(recID, Token, typeID*, flookup*);
extern Token	i_trampoline_func	(flookup*, bool[]);

extern void	alloc_and_init		(OUTSTREAM, recID, Token, Token, Token);
extern void	alloc_and_init_dcl	(OUTSTREAM, recID, Token, bool);

//*****************************************************************************
// breakexpr.h
//*****************************************************************************

typedef enum {
	NOAMBIG,
	PARENTH_CAST, PARENTH_SYNTAX, PARENTH_FCALL, PARENTH_POSTFIX,
	SIZEOF_TYPE, SIZEOF_EXPRESSION,
} exambg;

typedef struct {
	Token *expr;
	exambg et;
	NormPtr operators [128];
	int nop;
	bool esc_op;

	Token **branch;
} bexpr;

extern void	break_expr	(bexpr*);

enum { CONST_EXPR, INIT_EXPR, NORMAL_EXPR };

extern NormPtr	skip_expression	(Token*, NormPtr, int);

//*****************************************************************************
// rexpr.h
//*****************************************************************************

extern struct obflag {
	bool yes;
	recID class;
	bool have_this;
	typeID classP;
	Token func, efunc;
	bool polymorph, recording;
	bool isdtor;
} objective;

typedef struct {
	Token *newExpr;
	bool isconstant, ok;
	typeID type;
} exprret;

extern NormPtr	parse_expression		(NormPtr, exprret*, int);
extern NormPtr	parse_expression_retconv	(NormPtr, exprret*, typeID, int);
extern NormPtr	parse_const_expression		(NormPtr, exprret*);

extern void	rewrite_designator	(typeID, Token[]);
extern Token	*rewrite_ctor_expr	(Token*);

extern inline bool	isstructure	(typeID);
extern inline bool	ispointer	(typeID);
extern inline bool	isstructptr	(typeID);

extern bool	is_object_in_scope	(Token);

//*****************************************************************************
// statement.h
//*****************************************************************************

extern NormPtr	compound_statement		(OUTSTREAM, NormPtr);
extern NormPtr	statement			(OUTSTREAM, NormPtr);
extern Token*	rewrite_compound_statement	(Token*);

//*****************************************************************************
// dcl.h
//*****************************************************************************

extern bool	is_array_of_ctorable_objects	(typeID);

extern NormPtr	gen_array_ctors		(NormPtr, typeID, Token, int*, Token**);
extern NormPtr	local_declaration	(OUTSTREAM, NormPtr);
extern void	reparse_template_func	(Token, recID, NormPtr);
extern typeID	eval_cast		(Token**);
extern void	translation_unit	();
extern Token	bt_macro, bt_replace;

extern recID	current_scope[];
extern int	top_scope;

//*****************************************************************************
// fdb.h
//*****************************************************************************

typedef enum { DT_NORM, DT_AUTO, DT_ABSTRACT } deftype;

extern typeID	return_typeID;
extern void	store_definition
	(Token, Token*, Token*, typeID*, NormPtr, recID, typeID, Token, bool, deftype, bool);
extern void	store_define_dtor	(Token, recID);
extern void	commit_auto_define	(Token, recID, Token, Token);
extern void	define_auto_functions	();
extern void	commit_abstract_define	(Token, Token, recID, bool);
extern void	rename_fdb		(Token, Token);
extern void	remove_struct_from_def	(Token);
extern void	do_functions		();

// -- usage

extern void usage_tconst	(Token);
extern void usage_upcast	(recID);
extern void usage_fcall		(Token);
extern void usage_memb		(Token);
extern void usage_vvar		(Token*);
extern void usage_notok		();

//*****************************************************************************
// except.h
//*****************************************************************************

extern NormPtr throw_statement	(OUTSTREAM, NormPtr);
extern NormPtr try_statement	(OUTSTREAM, NormPtr);
extern void init_except		();
extern void decl_except_data	();

//*****************************************************************************
// regexp.h
//*****************************************************************************

extern Token	perlop_regexp	(Token);
extern NormPtr	parse_RegExp	(NormPtr, int);

//*****************************************************************************
// test.h
//*****************************************************************************

#define SAVE_VAR(x, y) __typeof__(x) _tmp_ ## x = x; x = y
#define RESTOR_VAR(x) x = _tmp_ ## x

#define SAVE_VARC(x, y, tmp) __typeof__(x) tmp = x; x = y
#define RESTOR_VARC(x, tmp) x = tmp;

#define SAVE_CODE(x) SAVE_VAR (CODE, x); SAVE_VAR (last_location, 0);
#define RESTOR_CODE  RESTOR_VAR (CODE); RESTOR_VAR (last_location);

static inline int max (int i, int j) { return i > j ? i : j; }
static inline int min (int i, int j) { return i < j ? i : j; }
static inline int in2 (int a, int b, int c) { return a == b || a == c; }
static inline int xor (int a, int b) { return a != 0 ^ b != 0; }
static inline int nz (int a) { return a != 0; }
