#include "global.h"

NormPtr statement (OUTSTREAM, NormPtr);

static NormPtr parenth_expression (OUTSTREAM o, NormPtr p)
{
	exprret E;
	if (CODE [p++] != '(')
		parse_error (p, "missing '(' after reserved keyword");
	p = parse_expression (p, &E, NORMAL_EXPR);
	if (!E.ok) parse_error (p, "compilation stops here");
	if (CODE [p++] != ')')
		parse_error (p, "missing ')' after reserved keyword");
	outprintf (o, '(', ISTR (E.newExpr), ')', -1);
	free (E.newExpr);
	return p;
}

static NormPtr __asm___statement (OUTSTREAM o, NormPtr p)
{
	NormPtr p2 = p;
	if (CODE [p] == RESERVED_volatile)
		++p;
	if (CODE [p++] != '(')
		parse_error (p, "__asm__ '('");
	p = skip_parenthesis (p);
	if (CODE [p++] != ';')
		parse_error (p, "missing ';' after __asm__");
	outprintf (o, RESERVED___asm__, NSTRNEXT, CODE + p2, p - p2, ';', - 1);
	return p;
}

static NormPtr if_statement (OUTSTREAM o, NormPtr p)
{
	output_itoken (o, RESERVED_if);
	p = statement (o, parenth_expression (o, p));
	// oh my god. this must be the horrible if-else ambiguity!
	if (CODE [p] == RESERVED_else) {
		output_itoken (o, RESERVED_else);
		p = statement (o, p + 1);
	}
	return p;
}

static NormPtr switch_statement (OUTSTREAM o, NormPtr p)
{
	Token swval = -1;
	/* Feature: switch [value] */
	if (CODE [p] == '[') {
		if (!ISSYMBOL (CODE [p + 1]) || CODE [p + 2] != ']')
			parse_error (p, "switch [identifier]");
		swval = CODE [p + 1];
		open_local_scope ();
		enter_local_obj (swval, typeID_int);
		outprintf (o, '{', RESERVED_int, swval, ';',
			   RESERVED_switch, '(', swval, '=', -1);
		p = parenth_expression (o, p + 3);
		output_itoken (o, ')');
	} else {
	/* ^^^^^^^^^^^^^^^^^^^^^^^ */
		output_itoken (o, RESERVED_switch);
		p = parenth_expression (o, p);
	}

	add_catchpoint (1);
	p = statement (o, p);
	rmv_catchpoint ();

	if (swval != -1) {
		close_local_scope ();
		output_itoken (o, '}');
	}
	return p;
}

static NormPtr do_statement (OUTSTREAM o, NormPtr p)
{
	OUTSTREAM S = new_stream ();

	add_catchpoint (2);
	p = statement (S, p);
	rmv_catchpoint ();

	if (CODE [p] == RESERVED_while) {
		output_itoken (o, RESERVED_do);
		concate_streams (o, S);
		output_itoken (o, RESERVED_while);
		p = parenth_expression (o, p + 1);
		if (CODE [p++] != ';')
			parse_error (p, "this is special. You need a ';'");
		output_itoken (o, ';');
	/* Feature: do without while */
	} else {
		outprintf (o, RESERVED_for, '(', ';', ';', ')', '{', - 1);
		concate_streams (o, S);
		outprintf (o, RESERVED_break, ';', '}', - 1);
	}
	/* ^^^^^^^^^^^^^^^^^^^^^^^^^ */

	return p;
}

static NormPtr for_statement (OUTSTREAM o, NormPtr p)
{
	bool first_dcl;
	exprret E1, E2, E3;

	if (CODE [p++] != '(')
		parse_error (p, "for '('");
	/* Feature: first part may be declaration */
	if (first_dcl = is_dcl_start (CODE [p])) {
		outprintf (o, '{', -1);
		open_local_scope ();
		p = local_declaration (o, p) - 1;
		(E1.newExpr = mallocint (1)) [0] = -1;
		E1.ok = true;
	} else p = parse_expression (p, &E1, NORMAL_EXPR);
	if (CODE [p++] != ';')
		parse_error (p, "for (';'");
	p = parse_expression (p, &E2, NORMAL_EXPR);
	if (CODE [p++] != ';')
		parse_error (p, "for (;';'");
	p = parse_expression (p, &E3, NORMAL_EXPR);
	if (CODE [p++] != ')')
		parse_error (p, "for (;';')");
	if (!(E1.ok && E2.ok && E3.ok))
		parse_error (p, "compilation halted at for()");
	outprintf (o, RESERVED_for, '(', ISTR (E1.newExpr), ';',
		   ISTR (E2.newExpr), ';', ISTR (E3.newExpr), ')', -1);
	free (E1.newExpr);
	free (E2.newExpr);
	free (E3.newExpr);
	add_catchpoint (2);
	p = statement (o, p);
	rmv_catchpoint ();
	if (first_dcl) {
		close_local_scope ();
		outprintf (o, '}', -1);
	}
	return p;
}

static NormPtr while_statement  (OUTSTREAM o, NormPtr p)
{
	output_itoken (o, RESERVED_while);
	add_catchpoint (2);
	p = statement (o, parenth_expression (o, p));
	rmv_catchpoint ();
	return p;
}

static NormPtr return_statement (OUTSTREAM o, NormPtr p)
{
	exprret E;
	bool he;
	p = parse_expression_retconv (p, &E, return_typeID, NORMAL_EXPR);
	if (!E.ok) return p + 1;

	/* Feature: a 'return;' in destructors, returns this */
	if (!(he = E.newExpr [0] != -1) && (he = objective.isdtor))
		sintprintf (frealloc (&E.newExpr, 2), RESERVED_this, -1);

	/* Feature: destructors for all locals */
	if (func_has_dtors ()) {
		outprintf (o, '{', -1);
		if (he) outprintf (o, RESERVED___typeof__, '(', ISTR (E.newExpr), ')',
			   internal_identifier1 (), '=', ISTR (E.newExpr), ';', -1);
		gen_all_destructors (o);
		outprintf (o, RESERVED_return, he ? internal_identifier1 () : BLANKT, ';', '}', -1);
	} else
	/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
		outprintf (o, RESERVED_return, ISTR (E.newExpr), ';', -1);
	free (E.newExpr);
	if (CODE [p] != ';')
		parse_error (p, "expecting ';' after return");
	return p + 1;
}

static NormPtr case_statement (OUTSTREAM o, NormPtr p)
{
	exprret E;
	output_itoken (o, RESERVED_case);
	p = parse_expression (p, &E, NORMAL_EXPR);
	if (E.ok) outprintf (o, ISTR (E.newExpr), -1);
	free (E.newExpr);
	/* extension '...' in case */
	if (CODE [p] == ELLIPSIS) {
		p = parse_expression (p, &E, NORMAL_EXPR);
		if (E.ok) {
			outprintf (o, ELLIPSIS, ISTR (E.newExpr), -1);
			free (E.newExpr);
		}
	}
	if (CODE [p] != ':') parse_error (p, "case ':' error");
	output_itoken (o, ':');
	return p + 1;
}

static NormPtr expression_statement (OUTSTREAM o, NormPtr p)
{
	exprret E;
	p = parse_expression (p, &E, NORMAL_EXPR);
	if (CODE [p++] != ';') parse_error (p, "expression';'");
	if (E.ok) {
		outprintf (o, ISTR (E.newExpr), ';', -1);
		free (E.newExpr);
	}
	return p;
}

NormPtr compound_statement (OUTSTREAM o, NormPtr p)
{
	open_local_scope ();
	output_itoken (o, '{');
	while (CODE [p] != '}')			/* this below is <classname>.<virtual-var> */
		p = is_dcl_start (CODE [p]) && !(ISSYMBOL (CODE [p]) && CODE [p + 1] == '.') ?
			local_declaration (o, p) : statement (o, p);
	/* Feature: destructors of local objects */
	if (scope_has_dtors ())
		gen_auto_destruction (o);
	/* ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ ~ */
	output_itoken (o, '}');
	close_local_scope ();
	return p + 1;
}

static NormPtr jump_statement (OUTSTREAM o, NormPtr p, Token t)
{
	int cont = t == RESERVED_continue ? 2 : 1;

	if (CODE [p] != ';')
		parse_error (p, "break/continue ';'");

	/* Feature: destructors of local objects */
	if (break_has_dtors (cont)) {
		output_itoken (o, '{');
		gen_break_destructors (o, cont);
		outprintf (o, t, ';', '}', -1);
	} else
	/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */
		outprintf (o, t, ';', -1);
	return p + 1;
}

NormPtr statement (OUTSTREAM o, NormPtr p)
{
	Token t = CODE [p++];

	if (ISSYMBOL (t) && CODE [p] == ':') {
		outprintf (o, t, CODE [p++], -1);
		return CODE [p] == '}' ? p : statement (o, p);
	}

	switch (t) {
	case '{':		return compound_statement (o, p);
	case RESERVED_case:	return case_statement (o, p);
	case RESERVED_return:	return return_statement (o, p);
	case RESERVED_while:	return while_statement (o, p);
	case RESERVED_for:	return for_statement (o, p);
	case RESERVED_do:	return do_statement (o, p);
	case RESERVED_switch:	return switch_statement (o, p);
	case RESERVED_if:	return if_statement (o, p);
	case RESERVED_else:	parse_error (p, "else without if");
	case RESERVED_throw:	return throw_statement (o, p);
	case RESERVED_try:	return try_statement (o, p);
	case RESERVED___asm__:	return __asm___statement (o, p);
	case RESERVED_continue:
	case RESERVED_break:	return jump_statement (o, p, t);
	default:		return expression_statement (o, p - 1);
	case ';':
		output_itoken (o, ';');
		return p;
	case RESERVED_goto:
		if (!ISSYMBOL (CODE [p]) || CODE [p + 1] != ';')
			parse_error (p, "'goto label;' only");
		outprintf (o, RESERVED_goto, CODE [p], ';', -1);
		return p + 2;
	case RESERVED_default:
		if (CODE [p] != ':') parse_error (p, "default:");
		outprintf (o, RESERVED_default, ':', -1);
		return p + 1;
	}
}

/* This is a compound statement inside an expression */
Token *rewrite_compound_statement (Token *t)
{
	OUTSTREAM o = new_stream ();

	SAVE_VAR (CODE, t);
	SAVE_VAR (last_location, 0);

	output_itoken (o, '(');
	compound_statement (o, 0);
	output_itoken (o, ')');

	RESTOR_VAR (CODE);
	RESTOR_VAR (last_location);

	return combine_output (o);
}
